"""
The Model-Context Protocol (MCP) module for Pentest-Tools.com. It uses the ptt API to provide tools, resources and prompts
to LLM agents that implement MCP.
"""

try:
    from mcp.server.fastmcp import Context, FastMCP
    from fastmcp.utilities import json_schema_type
    from mcp.types import SamplingMessage, TextContent
    from fastmcp.server.auth import TokenVerifier, AccessToken
    from fastmcp.server.dependencies import get_access_token
    from mcp.server.auth.settings import AuthSettings

    # TODO: It might be a good idea to use fastmcp2 here
    # from fastmcp import Context, FastMCP
except ImportError:
    # We use this as a guard, in case `pentesttools` was not installed with mcp support
    FastMCP = None


import asyncio
from copy import deepcopy
import json
import os
import sys
import time
import logging
from typing import Any, Literal

import pentesttools as ptt
from pentesttools.api import Finding

from pydantic import BaseModel, Field


class ConfirmDeleteSchema(BaseModel):
    confirm: str = Field(
        ...,
        description="Type 'yes' to confirm deletion, 'no' to cancel",
        pattern="^(yes|no)$",
    )


# JSON schema for the FindSubdomainsParams type
FIND_SUBDOMAINS_SEARCH_METHODS_SCHEMA = {
    "type": "object",
    "description": "Methods to be used for discovery. All are used by default for a deep scan. For a custom scan, you need to specify at least one.",
    "properties": {
        "passive_detection": {
            "type": "boolean",
            "default": True,
            "description": "Search historical subdomains",
        },
        "dns_records": {
            "type": "boolean",
            "default": True,
            "description": "Zone transfer",
        },
        "dns_enumeration": {
            "type": "object",
            "properties": {
                "enabled": {
                    "type": "boolean",
                    "default": True,
                },
                "wordlist_id": {
                    "type": "integer",
                    "default": 1600,
                },
            },
            "description": "DNS bruteforcing using a wordlist",
        },
        "ctr_search": {
            "type": "boolean",
            "default": True,
            "description": "Certificate Transparency Logs",
        },
        "external_api_search": {
            "type": "boolean",
            "default": True,
            "description": "Subdomains from external APIs",
        },
        "bing_search": {
            "type": "boolean",
            "default": True,
            "description": "Bing search",
        },
        "google_search": {
            "type": "boolean",
            "default": True,
            "description": "Google search",
        },
        "html_search": {
            "type": "boolean",
            "default": True,
            "description": "Subdomains extracted from HTML links and HTTP headers",
        },
        "ssl_search": {
            "type": "boolean",
            "default": True,
            "description": "Extract from SSL certificates",
        },
        "revdns_search": {
            "type": "boolean",
            "default": True,
            "description": "Reverse DNS on target IP ranges",
        },
        "alteration_search": {
            "type": "boolean",
            "default": True,
            "description": (
                "Generates permutations and alterations of the subdomain names found so far in the scan. "
                "Disabling this option should speed up scans when there is a large number of subdomains, but might reduce the number of results."
            ),
        },
        "cname_search": {
            "type": "boolean",
            "default": True,
            "description": "Extract from CNAMEs",
        },
    },
}
FindSubdomainsSearchMethods = json_schema_type.json_schema_to_type(FIND_SUBDOMAINS_SEARCH_METHODS_SCHEMA)


# JSON schema for the Discovery type
DISCOVERY_SCHEMA = {
    "type": "array",
    "items": {
        "type": "string",
        "enum": [
            "fingerprint",
            "software_vulnerabilities",
            "check_robots",
            "outdated_js",
            "untrusted_certificates",
            "client_access_policies",
            "http_debug_methods",
            "security_txt",
            "cors_misconfiguration",
            "resource_discovery",
            "sensitive_files",
            "admin_consoles",
            "interesting_files",
            "server_info_disc",
            "server_software",
            "misconfigurations",
            "graphql_endpoint",
            "fuzz_openapi_locations",
        ],
    },
    "description": "Discovery modules to run. If not set, all will run. If empty list, none will run.",
}

ATTACK_PASSIVE_SCHEMA = {
    "type": "array",
    "items": {
        "type": "string",
        "enum": [
            "security_headers",
            "cookie_security",
            "directory_listing",
            "secure_communication",
            "weak_password_submission",
            "error_debug_messages",
            "password_cleartext",
            "cross_domain_source",
            "mixed_content",
            "sensitive_data",
            "login_interfaces",
            "file_upload",
            "path_disclosure",
            "sql_statement_in_request",
            "password_in_response",
            "session_token_in_url",
            "api_endpoint",
        ],
    },
    "description": "Passive attacks to run. If not set, all will run. If empty list, none will run.",
}

# JSON schema for the AttackActive type
ATTACK_ACTIVE_SCHEMA = {
    "type": "array",
    "items": {
        "type": "string",
        "enum": [
            "xss",
            "sqli",
            "lfi",
            "oscmdi",
            "ssrf",
            "open_redirect",
            "broken_authentication",
            "php_code_injection",
            "js_code_injection",
            "ruby_code_injection",
            "python_code_injection",
            "perl_code_injection",
            "log4j_rce",
            "ssti",
            "xxe",
            "viewstate_rce",
            "prototype_pollution",
            "backup_files",
            "request_url_override",
            "http_request_smuggling",
            "session_fixation",
            "idor",
            "jwt",
            "response_header_injection",
        ],
    },
    "description": "Active attacks to run. If not set, all will run. If empty list, none will run.",
}

# JSON schema for the Spider type
SPIDER_SCHEMA = {
    "type": "object",
    "properties": {
        "approach": {
            "type": "string",
            "enum": ["auto", "classic", "spa"],
            "default": "auto",
        },
        "depth": {"type": "integer", "default": 10},
        "exclude_urls": {
            "type": "array",
            "items": {"type": "string"},
            "description": "URLs to exclude from spidering.",
        },
        "include_urls": {
            "type": "array",
            "items": {"type": "string"},
            "description": "URLs to include in spidering.",
        },
        "subdomains_in_scope": {
            "type": "array",
            "items": {"type": "string"},
            "description": "Subdomains to include in scope.",
        },
    },
}

Discovery = json_schema_type.json_schema_to_type(DISCOVERY_SCHEMA)
AttackPassive = json_schema_type.json_schema_to_type(ATTACK_PASSIVE_SCHEMA)
AttackActive = json_schema_type.json_schema_to_type(ATTACK_ACTIVE_SCHEMA)
Spider = json_schema_type.json_schema_to_type(SPIDER_SCHEMA)


NETWORK_SCANNER_PRESET_SCHEMA = {
    "type": "string",
    "enum": ["light", "deep", "custom"],
    "default": "deep",
    "description": (
        "Determines the Network Scanner scan type that will be used. "
        "`light` uses the `version_based` scanning engine on top 100 ports. "
        "`deep` uses the `version_based`, `sniper` and `nuclei` scanning engines on top 5000 ports (TCP) or top 1000 ports (UDP). "
        "`custom` allows you to use custom configurations of scanning engines and ports."
    ),
}
NetworkScannerPreset = json_schema_type.json_schema_to_type(NETWORK_SCANNER_PRESET_SCHEMA)

NETWORK_SCANNER_SCANNING_ENGINES = {
    "type": "array",
    "description": (
        "Scanning engines that will be loaded for a scan. Can be set only if `custom` preset was selected. "
        "It must contain at least one engine."
    ),
    "items": {
        "type": "string",
        "enum": ["version_based", "sniper", "nuclei", "openvas"],
    },
    "example": ["version_based", "sniper", "nuclei", "openvas"],
}
NetworkScannerScanningEngines = json_schema_type.json_schema_to_type(NETWORK_SCANNER_SCANNING_ENGINES)


NETWORK_SCANNER_PORT_RANGE = {
    "type": "object",
    "properties": {
        "start_port": {"type": "integer"},
        "end_port": {"type": "integer"},
    },
    "example": {"start_port": 0, "end_port": 443},
}
NetworkScannerPortRange = json_schema_type.json_schema_to_type(NETWORK_SCANNER_PORT_RANGE)

NETWORK_SCANNER_PORT_LIST = {
    "type": "array",
    "items": {
        "type": "integer",
        "minimum": 1,
        "maximum": 65535,
    },
    "example": [22, 80, 443],
}
NetworkScannerPortList = json_schema_type.json_schema_to_type(NETWORK_SCANNER_PORT_LIST)

NETWORK_SCANNER_ENGINE_OPTIONS = {
    "type": "object",
    "properties": {
        "requests_per_second": {
            "type": "integer",
            "default": 50,
            "description": (
                "Specify a maximum number of requests per second between 1 and 1000. "
                "It is only valid with the `deep` preset or with the `custom` preset if `nuclei` and/or `openvas` are included in `scanning_engines`."
            ),
            "minimum": 1,
            "maximum": 1000,
        },
        "cves_to_scan": {
            "type": "array",
            "default": [],
            "description": ("Specify particular CVEs for scanning. Can be set only if `custom` preset was selected."),
            "items": {"type": "string"},
            "example": ["CVE-2024-23108", "CVE-2024-27198"],
        },
    },
}
NetworkScannerEngineOptions = json_schema_type.json_schema_to_type(NETWORK_SCANNER_ENGINE_OPTIONS)


# TODO: This is a partial schema.
SCAN_PARAMS_SCHEMA = {
    "type": "object",
    "properties": {
        "workspace_id": {
            "type": "integer",
            "nullable": True,
            "writeOnly": True,
            "description": "Workspace where the scan is started. It has to match the workspace ID of the target",
        },
        "vpn_profile_uuid": {
            "type": "string",
            "nullable": True,
            "writeOnly": True,
            "description": "VPN profile to use for the scan. If null, there will be no VPN profile used. If not specified, the profile attached to the workspace will be used.",
        },
        "vpn_scan": {
            "type": "boolean",
            "readOnly": True,
            "description": "Indicates if the scan is performed over VPN. Read-only.",
        },
        "max_scan_time": {
            "type": "integer",
            "minimum": 1,
            "maximum": 1440,
            "default": 1440,
            "writeOnly": True,
            "description": (
                "Maximum number of minutes that the scan should run. "
                "Not supported by: Sniper, tools with short scan duration (like Website Recon or ICMP Ping)."
            ),
        },
    },
}

ScanParams = json_schema_type.json_schema_to_type(SCAN_PARAMS_SCHEMA)

os.environ["PTT_USER_AGENT"] = "ptt-cli-mcp"

API_URL = "https://app.pentest-tools.com/api/v2/"
MCP_METHOD = "stdio"
HEADERS = {
    "User-Agent": "ptt-cli-mcp",
    "Accept": "*/*",
}


class PTTApiKeyVerifier(TokenVerifier):
    """Verifies that the Bearer token is a valid Pentest-Tools API key."""

    async def verify_token(self, token: str) -> AccessToken | None:
        test_headers = {**HEADERS, "Authorization": f"Bearer {token}"}
        try:
            res = await asyncio.to_thread(ptt.api.get_workspaces, headers=test_headers)
            if res.status_code == 200:
                return AccessToken(token=token, client_id="ptt-user", scopes=[])
        except Exception:
            pass
        return None


def get_authentication_headers(context: Context) -> dict[str, str]:
    """A function that gives the headers for the request. We need this to get the PTT-API-KEY header

    Returns the headers necessary for a request, with the authentication added. If the key is not found, return an empty dict.
    """
    global HEADERS

    # For stdio, we get the api key from `ptt.api.api_key`
    if MCP_METHOD == "stdio":
        HEADERS["Authorization"] = f"Bearer {ptt.api.api_key}"
        return HEADERS

    # For http, we get the api key from the auth context (Bearer token = PTT API key)
    elif MCP_METHOD == "http":
        access_token = get_access_token()
        if access_token is not None:
            headers = deepcopy(HEADERS)
            headers["Authorization"] = f"Bearer {access_token.token}"
            return headers

    return {}


def run_tool(
    tool_id: int,
    target: str,
    tool_params: dict,
    scan_params: dict,
    context: Context,
    raw: bool = False,
):
    """The main way of running a scanning tool"""

    # `HEADERS` + authentication
    headers = get_authentication_headers(context)
    if not headers:
        return "The api key is missing."

    res = ptt.api.start_scan(target, tool_id, tool_params, scan_params, headers=headers)
    res_json = res.json()

    scan_id = res_json["data"]["created_id"]
    logging.info("Started scan %i, for tool_id %s", scan_id, tool_id)

    while True:
        time.sleep(2)

        # Get the status of our scan
        status = ptt.api.get_scan_status(scan_id, headers=headers)
        status_name = status.json()["data"]["status_name"]

        if status_name == "finished":
            logging.debug("Scan status: %s" % res_json["data"])
            logging.debug("Getting JSON report")

            if raw:
                res = ptt.api.get_raw_output(scan_id, headers=headers)
                return res.text

            res = ptt.api.get_output(scan_id, headers=headers)
            output_json = json.dumps(res.json())
            return output_json


async def run_tool_long_running(
    tool_id: int,
    target: str,
    tool_params: dict,
    scan_params: dict,
    context: Context,
    raw: bool = False,
):
    """
    Run a scanning tool.

    For light scans: runs to completion with progress tracking.
    For deep/custom scans: starts the scan and returns the scan_id immediately so the
    user can keep interacting with the LLM. Use the `scan` tool with verb="get_output"
    to retrieve results when ready.
    """

    headers = get_authentication_headers(context)

    if not headers:
        return "The api key is missing."

    res = ptt.api.start_scan(target, tool_id, tool_params, scan_params, headers=headers)
    res_json = res.json()

    try:
        scan_id = res_json["data"]["created_id"]
    except KeyError:
        return f"There was a problem starting the scan: {res_json}"

    logging.info("Started scan %i, for tool_id %s", scan_id, tool_id)

    # For deep/custom scans, return the scan_id immediately so the conversation is not blocked.
    # The user can retrieve results later using the `scan` tool with verb="get_output".
    if tool_params.get("scan_type", "light") != "light":
        return json.dumps({
            "scan_id": scan_id,
            "message": (
                f"Scan {scan_id} has been started. "
                "Use the `scan` tool with verb=\"get_output\" and "
                f"scan_id={scan_id} to retrieve results when ready, "
                "or verb=\"get_status\" to check progress."
                "You can always see the current running scans at https://app.pentest-tools.com/scans"
            ),
        })

    # Status groups
    progress_statuses = {"running", "waiting"}
    finished_statuses = {"finished"}
    failed_statuses = {
        "failed to start",
        "stopped",
        "timed out",
        "aborted",
        "VPN connection error",
        "auth failed",
        "connection error",
    }

    while True:
        await asyncio.sleep(2)

        # Get the status of our scan
        status = ptt.api.get_scan_status(scan_id, headers=headers)
        status_data = status.json()["data"]
        status_name = status_data["status_name"]

        # Report progress if running or waiting
        if context is not None and status_name in progress_statuses:
            await context.report_progress(
                float(status_data.get("progress", 0.0)),
                total=100.0,
                message="Scan id "
                + str(scan_id)
                + " | Status: "
                + str(status_name)
                + f" | Progress: {float(status_data.get('progress', 0.0))}",
            )

        if status_name in finished_statuses:
            logging.debug("Scan status: %s" % status_data)
            logging.debug("Getting JSON report")

            if raw:
                res = ptt.api.get_raw_output(scan_id, headers=headers)
                return res.text

            res = ptt.api.get_output(scan_id, headers=headers)
            output_json = json.dumps(res.json())
            return output_json

        elif status_name in failed_statuses:
            # Report failure to the LLM
            if context is not None:
                await context.info(
                    f"Scan {scan_id} failed with status: {status_name}. Status message: {status_data.get('status_message', '')}"
                )
            return json.dumps(
                {
                    "error": f"Scan failed with status: {status_name}",
                    "details": status_data,
                }
            )

        # Otherwise, keep waiting and reporting progress


_sentry_dsn = os.getenv("SENTRY_DSN")
if _sentry_dsn:
    try:
        import sentry_sdk
        from sentry_sdk.integrations.mcp import MCPIntegration

        sentry_sdk.init(
            dsn=_sentry_dsn,
            traces_sample_rate=1.0,
            integrations=[MCPIntegration()],
            send_default_pii=False,
        )
    except ImportError:
        pass  # sentry-sdk not installed, skip


if FastMCP is not None:
    mcp = FastMCP(
        "ptt-http",
        host=os.getenv("MCP_HOST", "127.0.0.1"),
        port=9000,
        token_verifier=PTTApiKeyVerifier(),
        auth=AuthSettings(
            issuer_url="https://app.pentest-tools.com",
            resource_server_url="http://localhost:9000",
        ),
    )

    ## Resources
    # TODO: json data may not be suitable for LLMs, at least for the less powerful ones.
    # Resources are generally the GET requests from API
    @mcp.resource("scans://{garbage}")
    def scans(garbage: str, context: Context) -> str:
        """Get the current scans (all, not just running) as JSON string."""
        headers = get_authentication_headers(context)
        if not headers:
            return "The api key was not set"

        res = ptt.api.get_scans(headers=headers)
        return json.dumps(res.json())

    @mcp.resource("targets://{garbage}")
    def targets(garbage: str, context: Context) -> str:
        """Get the current targets available"""
        headers = get_authentication_headers(context)
        if not headers:
            return "The api key was not set"

        res = ptt.api.get_targets(headers=headers)
        return json.dumps(res.json())

    @mcp.resource("workspaces://{garbage}")
    def workspaces(garbage: str, context: Context) -> str:
        """Get the current workspaces available"""
        headers = get_authentication_headers(context)
        if not headers:
            return "The api key was not set"

        res = ptt.api.get_workspaces(headers=headers)
        return json.dumps(res.json())

    @mcp.resource("http_loggers://{garbage}")
    def http_loggers(garbage: str, context: Context) -> str:
        """Get the current http_loggers available"""
        headers = get_authentication_headers(context)
        if not headers:
            return "The api key was not set"

        res = ptt.api.get_http_loggers(headers=headers)
        return json.dumps(res.json())

    @mcp.resource("wordlists://{garbage}")
    def wordlists(garbage: str, context: Context) -> str:
        """Get the current wordlists available"""
        headers = get_authentication_headers(context)
        if not headers:
            return "The api key was not set"

        res = ptt.api.get_wordlists(headers=headers)
        return json.dumps(res.json())

    @mcp.resource("findings://{garbage}")
    def findings(garbage: str, context: Context) -> str:
        """Get the current findings available"""
        headers = get_authentication_headers(context)
        if not headers:
            return "The api key was not set"

        res = ptt.api.get_findings(headers=headers)
        return json.dumps(res.json())

    @mcp.resource("reports://{garbage}")
    def reports(garbage: str, context: Context) -> str:
        """Get the current reports available"""
        headers = get_authentication_headers(context)
        if not headers:
            return "The api key was not set"

        res = ptt.api.get_all_reports(headers=headers)
        return json.dumps(res.json())

    ### Prompts
    # Prompts should match the intended use-cases for the pentesttools MCP
    @mcp.prompt()
    def pentest(target: str) -> str:
        """Create a pentesting report prompt"""
        return f"Please provide a comprehensive pentesting report for this target: {target}, using the tools at your disposal."

    @mcp.prompt()
    def is_up(target: str) -> str:
        """Create a simple target checking prompt"""
        return f"Please see if this target: {target} is up, using the tools at your disposal."

    @mcp.prompt()
    def subdomains(target: str) -> str:
        """Discover all subdomains for the target domain"""
        return (
            f"Please find all the subdomains that you can for this domain: {target}, using the tools at your disposal."
        )

    @mcp.prompt()
    def summary(workspace_id: int) -> str:
        """Create an executive summary for scans in a certain time period from workspace `workspace_id`"""
        return f"Please create an executive summary for the latest scans in the workspace {workspace_id}, using the tools at your disposal."

    @mcp.prompt()
    def vibeseccoding() -> str:
        """Run a vpn website scanner on the target and fix the problems"""

        return """
        You are inside an application. Please run a `pwd` command to check that you have terminal commands enabled.

        Please inspect the docker related files to see how the app is built and run.

        Run the application. In case you don't figure out how, try `docker compose build -d`, or other similar general commands.

        Use `docker network inspect mcp` in order to see what ip docker gave the application.

        Then please run a deep website scan on it and fix the most critical vulnerabilities you find.
        """

    def get_target_name_by_scan_id(scan_id: int, headers) -> str:
        """Get the target name for a given scan ID."""
        scan_res = ptt.api.get_scan_status(scan_id, headers=headers)
        target_id = scan_res.json()["data"]["target_id"]

        target_res = ptt.api.get_target_by_id(target_id, headers=headers)
        return target_res.json()["data"]["name"]


    def get_workspace_id_by_scan_id(scan_id: int, headers) -> int:
        """Get the workspace ID for a given scan ID."""
        scan_res = ptt.api.get_scan_status(scan_id, headers=headers)
        target_id = scan_res.json()["data"]["target_id"]

        target_res = ptt.api.get_target_by_id(target_id, headers=headers)
        return target_res.json()["data"]["workspace_id"]


    def get_findings_from_scans(scans, headers) -> tuple[int, str, list[str], str]:
        finding_ids = []
        target_name = ""
        workspace_id = 0
        # If scans are provided, get findings from each scan
        for scan_id in scans:
            if not target_name:
                try:
                    target_name = get_target_name_by_scan_id(scan_id, headers)
                except Exception:
                    logging.info("Cannot get target from scan_id, skipping this scan")
                    continue

            if not workspace_id:
                try:
                    workspace_id = get_workspace_id_by_scan_id(scan_id, headers)
                except Exception:
                    logging.info("Cannot get the workspace_id from scan_id, skipping")
                    continue

            scan_output = ptt.api.get_output(scan_id=scan_id, headers=headers).json()
            try:
                for finding in scan_output["data"]["output_data"]["findings"]:
                    finding_ids.append(finding["id"])
            except KeyError:
                continue

        if not finding_ids:
            return workspace_id, target_name, finding_ids, "No findings found in the provided scans."
        
        if not target_name:
            return workspace_id, target_name, finding_ids, "No target name could be extracted"

        return workspace_id, target_name, finding_ids, ""

    def get_findings_from_findings(findings, headers):
        if len(findings) == 0:
            return 0, "", [], "No valid finding IDs provided."

        error_message = ""
        target_name = ""
        workspace_id = 0

        first_finding = findings[0]

        finding_res = ptt.api.get_finding_information_by_id(finding_id=first_finding, headers=headers)
        finding_data = finding_res.json().get("data", {})
        if "target_name" in finding_data:
            target_name = finding_data["target_name"]

        target_id = finding_data.get("target_id")
        if target_id:
            try:
                target_res = ptt.api.get_target_by_id(target_id, headers=headers)
                workspace_id = target_res.json()["data"]["workspace_id"]
            except Exception:
                logging.info("Cannot get workspace_id from target_id %s", target_id)

        return workspace_id, target_name, findings, error_message

    def get_findings_from_json_report(report_id, headers):
        finding_ids = []
        error_message = ""
        target_name = ""
        workspace_id = 0

        headers["Accept"] = "application/json"
        report_json = []
        try:
            report_res = ptt.api.download_report(report_id, headers=headers)
            report_json = report_res.json()
        except Exception:
            error_message = "The report could not be downloaded as JSON. Please ensure that the target report is available in JSON format."
        # Extract finding_ids and target_name
        if isinstance(report_json, list) and report_json:
            for item in report_json:
                findings_list = item.get("findings", [])
                finding_ids.extend([f["id"] for f in findings_list if "id" in f])
            if report_json[0].get("findings"):
                target_name = report_json[0]["findings"][0].get("target_name", "")
            else:
                error_message = "Could not extract findings from the report."

        if finding_ids:
            try:
                finding_res = ptt.api.get_finding_information_by_id(finding_id=finding_ids[0], headers=headers)
                target_id = finding_res.json().get("data", {}).get("target_id")
                if target_id:
                    target_res = ptt.api.get_target_by_id(target_id, headers=headers)
                    workspace_id = target_res.json()["data"]["workspace_id"]
            except Exception:
                logging.info("Cannot get workspace_id from report findings")

        return workspace_id, target_name, finding_ids, error_message

    ### Tools
    # These are the actual pentesting tools, with some added management tools, available through `ptt.api`.
    # There are also some meta-tools that the MCP sees as tools as well.
    @mcp.tool()
    async def translate_report(
        report_id: int = None,
        findings: list[int] = None,
        scans: list[int] = None,
        target_language: str = None,
        workspace_id: int = None,  # TODO: We can't get this from the report, but can we get this from the findings themselves?
        context: Context = None,
    ) -> int | str:
        """Translate a report from English to the target language.

        IMPORTANT: This tool can only read and translate reports in JSON format.
        If you want to translate an existing report by report_id, that report must have
        been created with format="json". For PDF, HTML, or other formats, use `scans` or
        `findings` as input instead — the tool will create a new translated JSON report from
        those directly, without needing a pre-existing report.

        Args:
            report_id: The id of a JSON-format report to translate. Only one of `report_id`, `findings`, `scans` should be present.
            findings: The list of finding IDs to translate into a new report. Only one of `report_id`, `findings`, `scans` should be present.
            scans: The list of scan IDs to translate into a new report. Only one of `report_id`, `findings`, `scans` should be present.
            target_language: Target language to translate to, from English.
            workspace_id: The Workspace ID of the new findings that will be imported. The report is not associated with any workspace_id, so we need to pass one explicitly.

        Returns:
            The report_id of the new report.
        """
        headers = get_authentication_headers(context)
        if not headers:
            return "The api key was not set"

        # 0. Ensure only one of report_id, findings, or scans is present, and determine which one was chosen
        provided = [x is not None for x in (report_id, findings, scans)]
        if sum(provided) != 1:
            return "You must provide exactly one of: report_id, findings, or scans."

        if report_id is not None:
            chosen_input = "report_id"
        elif findings is not None:
            chosen_input = "findings"
        elif scans is not None:
            chosen_input = "scans"
        else:
            chosen_input = None  # Should not happen due to the check above

        # For report_id, verify it is a JSON report before proceeding
        if chosen_input == "report_id":
            info_res = ptt.api.get_report_information_by_id(report_id, headers=headers)
            report_info = info_res.json().get("data", {})
            report_format = report_info.get("format")
            if report_format != "json":
                return (
                    f"Report {report_id} is in '{report_format}' format. "
                    "translate_report can only read JSON reports. "
                    "Either create the report again with format='json', or pass the "
                    "underlying scan IDs via the `scans` parameter instead."
                )

        # Log the chosen input for debugging
        logging.info(f"Chosen input for translation: {chosen_input}")

        # 1. Get the finding_ids list and target_name based on the chosen input
        finding_ids: list[int] = []
        target_name = ""

        if chosen_input == "report_id":
            workspace_id, target_name, finding_ids, error_message = get_findings_from_json_report(report_id, headers)
        elif chosen_input == "findings":
            workspace_id, target_name, finding_ids, error_message = get_findings_from_findings(findings, headers)
        elif chosen_input == "scans":
            workspace_id, target_name, finding_ids, error_message = get_findings_from_scans(scans, headers)
        else:
            return "Invalid input for translation. Only one of `report_id`, `findings`, `scans` should be present."

        if error_message:
            logging.info(error_message)
            return error_message

        if not finding_ids:
            logging.info("No findings found.")
            return "No findings found."

        logging.info("finding_ids %s", finding_ids)

        # TODO: Translate these fields: risk_description, recommendation, name, vuln_description
        # 3. For each finding, get its details and elicit translation
        translated_findings: list[Finding] = []

        for fid in finding_ids:
            finding_res = ptt.api.get_finding_information_by_id(finding_id=fid, headers=headers)
            finding = finding_res.json().get("data", {})
            logging.info("Translating finding: %s", finding)

            # For now, we only translate the following fields:
            vuln_description = finding.get("vuln_description", "")
            vuln_recommendation = finding.get("recommendation", "")
            test_description = finding.get("test_description", "")
            finding_name = finding.get("name", "")

            # TODO: We have some problems when vuln_evidence should be a table here
            # It's still unclear if the api supports this
            vuln_evidence_obj = finding.get("vuln_evidence", {})
            if isinstance(vuln_evidence_obj, dict) and vuln_evidence_obj.get("type") == "text":
                vuln_evidence = vuln_evidence_obj.get("data", "")
            else:
                vuln_evidence = json.dumps(vuln_evidence_obj)

            logging.info("Will actually translate:")
            logging.info([vuln_description, vuln_recommendation, test_description, finding_name, vuln_evidence])

            # Do MCP Sampling for the translation for each field
            translated = {}
            for field, value in [
                ("vuln_description", vuln_description),
                ("vuln_recommendation", vuln_recommendation),
                ("test_description", test_description),
                ("finding_name", finding_name),
                ("vuln_evidence", vuln_evidence),
            ]:
                prompt = f"Translate the following {field} from English to {target_language}. Do not translate URLs, API endpoints, or vulnerability names unless appropriate.\n\n{value}"

                if value:
                    sampling_result = await context.session.create_message(
                        messages=[
                            SamplingMessage(
                                role="user",
                                content=TextContent(type="text", text=prompt),
                            )
                        ],
                        max_tokens=200,
                    )

                    if sampling_result.content.type == "text":
                        translated[field] = sampling_result.content.text
                    else:
                        translated[field] = value
                else:
                    translated[field] = value

            logging.info("Translated: %s", translated)

            # Prepare the finding to import
            finding_to_import = Finding(
                target_name=target_name,
                source=2,
                name=translated["finding_name"],
                test_description=translated["test_description"],
                vuln_description=translated["vuln_description"],
                vuln_risk_level=finding.get("vuln_risk_level"),
                vuln_recommendation=translated["vuln_recommendation"],
                cvss_v3=finding.get("cvss_v3"),
                vuln_evidence=translated["vuln_evidence"],
            )
            translated_findings.append(finding_to_import)

        # Import the translated finding (as a list)
        import_res = ptt.api.import_findings(
            findings=translated_findings,
            workspace_id=workspace_id,
            headers=headers,
        )
        import_data = import_res.json()

        logging.info("Imported data: %s", import_data)

        translated_finding_ids = []
        for item in import_data.get("data", {}).get("results", []):
            new_finding_id = item.get("id")
            if new_finding_id:
                translated_finding_ids.append(new_finding_id)

        if not translated_finding_ids:
            return "No translated findings were imported."

        logging.info(translated_finding_ids)

        # 4. Create a new report with the translated findings
        create_report_res = ptt.api.create_report(
            format="pdf",
            group_by="target",
            source="findings",
            resources=translated_finding_ids,
            headers=headers,
        )
        create_report_data = create_report_res.json()

        logging.info("Created report: %s", create_report_data)
        new_report_id = create_report_data.get("data", {}).get("report_id")
        if not new_report_id:
            return "Failed to create the translated report."

        return f"Download your new report from https://app.pentest-tools.com/reports/{new_report_id}"

    # Pentesting tools
    @mcp.tool()
    async def run_website_scanner(
        target: str,
        scan_type: Literal["light", "deep", "custom"] = "light",
        attack_active: "AttackActive" = None,
        attack_passive: "AttackPassive" = None,
        discovery: "Discovery" = None,
        spider: "Spider" = None,
        requests_per_second: int = 100,
        workspace_id: int = None,
        vpn_profile_uuid: str = None,
        auth_cookies: str = None,
        auth_headers: list[str] = None,
        auth_login_form_url: str = None,
        auth_login_form_username: str = None,
        auth_login_form_password: str = None,
        max_scan_time: int = 5,
        context: Context = None,
    ) -> str:
        """
        Run a website scanner on the target URL, reporting progress/status updates.

        Args:
            target: The target URL to scan.
            scan_type: The type of scan to perform. Possible values: "light", "deep", "custom". Default is "deep".
            attack_active: List of active attack modules to run.
            attack_passive: List of passive attack modules to run.
            discovery: List of discovery modules to run.
            spider: Spidering options as a dict.
            requests_per_second: Max requests per second (1-10000). Default is 100.
            workspace_id: The Workspace ID for the workspace the scan is going to be running in.
            vpn_profile_uuid: The VPN profile ID for the scan.
            auth_cookies: Cookie header string for authentication. Example: "PHPSESSID=abc123; token=xyz".
            auth_headers: List of HTTP headers for authentication. Example: ["Authorization: Bearer token123", "Cookie: session=abc"].
            auth_login_form_url: URL of the login page for form-based authentication.
            auth_login_form_username: Username for login form authentication.
            auth_login_form_password: Password for login form authentication.
            max_scan_time: The maximum duration in minutes for the scan. Between 5 and 1440 minutes.
        """
        tool_params = {
            "scan_type": scan_type,
            "requests_per_second": requests_per_second,
        }
        if attack_active:
            tool_params["attack_active"] = attack_active
        if attack_passive:
            tool_params["attack_passive"] = attack_passive
        if discovery:
            tool_params["discovery"] = discovery
        if spider:
            tool_params["spider"] = spider

        scan_params: dict[str, Any] = {"max_scan_time": str(max_scan_time)}
        if vpn_profile_uuid:
            scan_params["vpn_profile_uuid"] = vpn_profile_uuid

        if workspace_id:
            scan_params["workspace_id"] = workspace_id

        if auth_cookies:
            scan_params["authentication"] = {
                "type": "cookies",
                "params": {"cookie_header": auth_cookies},
            }
        elif auth_headers:
            scan_params["authentication"] = {
                "type": "headers",
                "params": {"headers": auth_headers},
            }
        elif auth_login_form_url and auth_login_form_username and auth_login_form_password:
            scan_params["authentication"] = {
                "type": "login_form",
                "params": {
                    "login_page_url": auth_login_form_url,
                    "username": auth_login_form_username,
                    "password": auth_login_form_password,
                },
            }
        elif any(
            (
                bool(auth_login_form_url),
                bool(auth_login_form_username),
                bool(auth_login_form_password),
            )
        ):
            return "You passed only one of `auth_login_form_url`, `auth_login_form_username`, `auth_login_form_password`, but you must pass all of them."

        return await run_tool_long_running(
            ptt.api.Tool.WEBSITE_SCANNER,
            target,
            tool_params,
            scan_params,
            context,
        )

    @mcp.tool()
    async def run_subdomain_finder(
        target: str,
        scan_type: Literal["light", "deep", "custom"] = "light",
        web_details: bool = None,
        whois: bool = None,
        unresolved_results: bool = None,
        search_methods: "FindSubdomainsSearchMethods" = None,
        max_scan_time: int = 5,
        context: Context = None,
    ) -> str:
        """
        Discover all subdomains for the target domain. Use to enumerate subdomains for further scanning.
        """
        tool_params = {
            "scan_type": scan_type,
        }

        if web_details is not None:
            tool_params["web_details"] = web_details
        if whois is not None:
            tool_params["whois"] = whois
        if unresolved_results is not None:
            tool_params["unresolved_results"] = unresolved_results
        if search_methods is not None:
            tool_params["search_methods"] = search_methods

        scan_params = {"max_scan_time": str(max_scan_time)}

        return await run_tool_long_running(
            ptt.api.Tool.SUBDOMAIN_FINDER,
            target,
            tool_params,
            scan_params,
            context,
        )

    @mcp.tool()
    async def run_network_scanner(
        target: str,
        preset: Literal["light", "deep", "custom"] = "light",
        scanning_engines: "NetworkScannerScanningEngines" = None,
        engines_options: "NetworkScannerEngineOptions" = None,
        check_alive: bool = None,  # True
        extensive_modules: bool = None,  # False
        protocol_type: Literal["tcp", "udp"] = "tcp",
        port_range: "NetworkScannerPortRange" = None,
        port_top: Literal["OpenVAS default", "10", "100", "1000", "5000"] = None,
        port_list: "NetworkScannerPortList" = None,
        just_scan_ports: bool = None,
        workspace_id: int = None,
        max_scan_time: int = 5,
        context: Context = None,
    ) -> str:
        """Run a network scanner on the target hostname.

        `just_scan_ports` will only scan the selected ports to see which are open, nothing heavy. It's also called "`port_scanner`".

        You can set port top only in custom mode.
        The scanning engines field is required when preset is custom.
        """

        tool_params = {"scan_type": preset}

        if just_scan_ports:
            # Set tool params
            if port_range:
                tool_params["port_range"] = port_range
            if port_top:
                tool_params["port_top"] = port_top
            if port_list:
                tool_params["port_list"] = port_list

            # Set scan params
            scan_params: dict[str, str | int] = {"max_scan_time": str(max_scan_time)}
            if workspace_id is not None:
                scan_params["workspace_id"] = workspace_id

            return await run_tool_long_running(
                ptt.api.Tool.PORT_SCANNER,
                target,
                tool_params,
                scan_params,
                context,
            )

        # Set tool params
        if check_alive is not None:
            tool_params["check_alive"] = check_alive
        if extensive_modules is not None:
            tool_params["extensive_modules"] = extensive_modules
        if protocol_type is not None:
            tool_params["protocol_type"] = protocol_type
        if scanning_engines:
            tool_params["scanning_engines"] = scanning_engines
        if engines_options:
            tool_params["engines_options"] = engines_options
        if port_range:
            tool_params["port_range"] = port_range
        if port_top:
            tool_params["port_top"] = port_top
        if port_list:
            tool_params["port_list"] = port_list

        # Set scan params
        scan_params = {"max_scan_time": str(max_scan_time)}
        if workspace_id is not None:
            scan_params["workspace_id"] = workspace_id

        return await run_tool_long_running(
            ptt.api.Tool.NETWORK_SCANNER,
            target,
            tool_params,
            scan_params,
            context,
        )

    # Administrative tools
    @mcp.tool()
    async def get_scans(
        context: Context,
        workspace_id: int = None,
        target_id: int = None,
        status: Literal[
            "running",
            "finished",
            "failed to start",
            "stopped",
            "timed out",
            "waiting",
            "aborted",
            "VPN connection error",
            "auth failed",
            "connection error",
        ] = None,
        limit: int = None,
        page: int = None,
        start_time: dict = None,
    ) -> str:
        """
        Retrieve scans from Pentest-Tools.com with optional filtering.

        Usage:
        - To get all scans: call with no extra parameters.
        - To filter by workspace: set workspace_id=<workspace id>.
        - To filter by target: set target_id=<target id>.
        - To filter by scan status: set status to one of the allowed values.
        - To limit results: set limit=<max number of scans>.
        - To paginate: set page=<page number>.
        - To filter by start time: set start_time as a dictionary (e.g. {"from": "...", "to": "..."}).

        Args:
            context: The MCP context object.
            workspace_id: (optional) ID of the workspace to show scans from.
            target_id: (optional) ID of the target to show scans from.
            status: (optional) Status of the scans to be shown. Allowed values: running, finished, failed to start, stopped, timed out, waiting, aborted, VPN connection error, auth failed, connection error.
            limit: (optional) Maximum number of scans to return.
            page: (optional) Page number to return.
            start_time: (optional) Dictionary specifying start time interval for scans (e.g. {"from": "...", "to": "..."}).

        Returns:
            JSON string containing the list of scans matching the filters.
        """
        headers = get_authentication_headers(context)
        if not headers:
            return "The api key was not set"
        res = ptt.api.get_scans(
            workspace_id=workspace_id,
            target_id=target_id,
            status=status,
            limit=limit,
            page=page,
            start_time=start_time,
            headers=headers,
        )
        return json.dumps(res.json())

    @mcp.tool()
    async def scan(
        verb: Literal[
            "get_status",
            "get_raw_output",
            "get_output",
            "start_by_targetid",
            "stop",
            "delete",
            "get_scan_output_by_id",
        ],
        context: Context,
        scan_id: int = None,
        target_id: int = None,
        tool_id: int = None,
        tool_params: dict = None,
    ) -> str:
        """
        Manage scans in Pentest-Tools.com.

        Usage:
        - To get the status of a scan: verb="get_status", scan_id=<scan id>
        - To get the raw output of a scan: verb="get_raw_output", scan_id=<scan id>
        - To get the parsed output of a scan: verb="get_output", scan_id=<scan id>
        - To start a scan by target id: verb="start_by_targetid", target_id=<target id>, tool_id=<tool id>, tool_params=<tool params>
        - To stop a scan: verb="stop", scan_id=<scan id>
        - To delete a scan: verb="delete", scan_id=<scan id>
        - To get a scan output of a scan: verb="get_scan_output_by_id", scan_id=<scan id>

        Only set the parameters required for the verb you use.

        Args:
            verb: The action to perform. Must be one of: get_status, get_raw_output, get_output, start_by_targetid, stop, delete.
            scan_id: The scan ID (required for get_status, get_raw_output, get_output, stop, delete).
            target_id: The target ID (required for start_by_targetid).
            tool_id: The tool ID (required for start_by_targetid).
            tool_params: The tool parameters (required for start_by_targetid).
        """
        headers = get_authentication_headers(context)
        if not headers:
            return "The api key was not set"

        if verb == "get_status":
            if not scan_id:
                return "Missing scan_id for get_status."
            res = ptt.api.get_scan_status(scan_id, headers=headers)
            return json.dumps(res.json())
        elif verb == "get_raw_output":
            if not scan_id:
                return "Missing scan_id for get_raw_output."
            res = ptt.api.get_raw_output(scan_id, headers=headers)
            return res.text
        elif verb == "get_output":
            if not scan_id:
                return "Missing scan_id for get_output."
            res = ptt.api.get_output(scan_id, headers=headers)
            return json.dumps(res.json())
        elif verb == "start_by_targetid":
            if not target_id or not tool_id or tool_params is None:
                return "Missing target_id, tool_id, or tool_params for start_by_targetid."
            res = ptt.api.start_scan_by_targetid(target_id, tool_id, tool_params, headers=headers)
            return json.dumps(res.json())
        elif verb == "stop":
            if not scan_id:
                return "Missing scan_id for stop."
            res = ptt.api.stop_scan(scan_id, headers=headers)
            return json.dumps(res.json())
        elif verb == "delete":
            if not scan_id:
                return "Missing scan_id for delete."

            # Elicit confirmation from the user before deleting
            if context is not None:
                result = await context.elicit(
                    message="Deleting a scan is an irreversible operation. Respond to this message in order to proceed",
                    schema=ConfirmDeleteSchema,
                )
                # Only proceed if user typed 'yes'
                if result.action != "accept":
                    return "Scan deletion cancelled by user."

            res = ptt.api.delete_scan(scan_id, headers=headers)
            return json.dumps(res.json())
        elif verb == "get_scan_output_by_id":
            if not scan_id:
                return "Missing scan_id for get_scan_output_by_id."
            res = ptt.api.get_output(scan_id, headers=headers)
            return json.dumps(res.json())
        else:
            return "Unknown verb: {verb}. Use one of: get_status, get_raw_output, get_output, start_by_targetid, stop, delete."

    # VPNs
    @mcp.tool()
    async def get_vpn_profiles(
        context: Context,
    ) -> str:
        """
        Use this tool to get a list of vpn profiles from Pentest-Tools.com.

        Returns:
            JSON string containing the list of targets matching the filters.
        """
        headers = get_authentication_headers(context)
        if not headers:
            return "The api key was not set"

        res = ptt.api.get_vpn_profiles(
            headers=headers,
        )
        return json.dumps(res.json())

    ### Targets
    @mcp.tool()
    async def get_targets(
        context: Context,
        workspace_id: int = None,
        target_type: Literal["hostname", "ip_address", "url"] = None,
        limit: int = None,
        page: int = None,
    ) -> str:
        """
        Use this tool to get a list of targets from Pentest-Tools.com.

        Args:
            context: The MCP context object.
            workspace_id: (optional) ID of the workspace to show targets from.
            target_type: (optional) The type of target. Possible values: "hostname", "ip_address", "url".
            limit: (optional) The maximum number of targets to return.
            page: (optional) The page number to return.

        Returns:
            JSON string containing the list of targets matching the filters.
        """
        headers = get_authentication_headers(context)
        if not headers:
            return "The api key was not set"
        res = ptt.api.get_targets(
            workspace_id=workspace_id,
            target_type=target_type,
            limit=limit,
            page=page,
            headers=headers,
        )
        return json.dumps(res.json())

    @mcp.tool()
    async def target(
        verb: Literal["add", "delete"],
        context: Context,
        name: str = None,
        description: str = "",
        workspace_id: int = None,
        target_id: int = None,
    ) -> str:
        """
        Manage targets in Pentest-Tools.com.

        Usage:
        - To add a new target: verb="add", name=<target name>, description=<optional>, workspace_id=<optional>.
        - To delete a target: verb="delete", target_id=<target id>.

        Only set the parameters required for the verb you use.

        Args:
            verb: The action to perform. Must be one of: add, delete.
            context: The MCP context object.
            name: The name of the target to add (required for add).
            description: Description for the target (optional for add).
            workspace_id: Workspace ID to associate with the target (optional for add).
            target_id: The ID of the target to delete (required for delete).
        """
        headers = get_authentication_headers(context)
        if not headers:
            return "The api key was not set"

        if verb == "add":
            if not name:
                return "Missing target name for add."
            res = ptt.api.add_target(name, description, workspace_id, headers=headers)
            return json.dumps(res.json())
        elif verb == "delete":
            if not target_id:
                return "Missing target_id for delete."
            res = ptt.api.delete_target(target_id, headers=headers)
            return json.dumps(res.json())
        else:
            return f"Unknown verb: {verb}. Use one of: add, delete."

    ### Reports
    @mcp.tool()
    async def create_report(
        format: Literal["pdf", "html", "json", "csv", "xlsx", "docx"],
        group_by: Literal["target", "vulnerability"],
        source: Literal["scans", "findings"],
        resources: list[int],
        context: Context,
        filters: dict = None,
        webhook_url: str = None,
    ) -> str:
        """Create a pentesting report, based on the findings already present in the platform.

        Args:
            format: The format in which to create the report. Possible values: pdf, html, json, csv, xlsx, docx.
            group_by: What key is used to create the report. Possible values: target, vulnerability.
            source: This is the source that is used to create the report. Possible values: scans, findings.
            resources: A list of ids, depeding on which source was chosen. If `scans` was chosed, then it's a list of scan_ids. If `findings` was chosed, then it's a list of finding_ids.
        """
        headers = get_authentication_headers(context)
        if not headers:
            return "The api key was not set"
        res = ptt.api.create_report(format, group_by, source, resources, filters, webhook_url, headers=headers)
        return json.dumps(res.json())

    @mcp.tool()
    async def report(
        verb: Literal["delete", "download", "get_all", "get_report_information_by_id"],
        context: Context,
        report_id: int = None,
    ) -> str:
        """
        Manage reports in Pentest-Tools.com.

        Usage:
        - To delete a report: verb="delete", report_id=<report id>
        - To download a report: verb="download", report_id=<report id>
        - To get all reports: verb="get_all"
        - To get report information: verb="get_report_information_by_id", report_id=<report id>

        Only set the parameters required for the verb you use.

        Args:
            verb: The action to perform. Must be one of: delete, download, get_all.
            context: The MCP context object.
            report_id: The ID of the report (required for delete and download).
        """
        headers = get_authentication_headers(context)
        if not headers:
            return "The api key was not set"

        if verb == "delete":
            if not report_id:
                return "Missing report_id for delete."
            res = ptt.api.delete_report(report_id, headers=headers)
            return json.dumps(res.json())
        elif verb == "download":
            if not report_id:
                return "Missing report_id for download."
            res = ptt.api.download_report(report_id, headers=headers)
            try:
                return res.json()
            except Exception:
                return res.text

        elif verb == "get_all":
            res = ptt.api.get_all_reports(headers=headers)
            return json.dumps(res.json())
        elif verb == "get_report_information_by_id":
            if not report_id:
                return "Missing report_id for get_report_information_by_id."
            res = ptt.api.get_report_information_by_id(report_id, headers=headers)
            return json.dumps(res.json())
        else:
            return f"Unknown verb: {verb}. Use one of: delete, download, get_all."

    ### Loggers
    @mcp.tool()
    async def http_logger(
        verb: Literal["create", "delete", "clear"],
        context: Context,
        label: str = None,
        logger_id: int = None,
        workspace_id: int = None,
    ) -> str:
        """
        Manage HTTP request loggers.

        Usage:
        - To create a new HTTP logger: call with verb="create", label="logger label", workspace_id=<optional workspace id>.
        - To delete an HTTP logger: call with verb="delete", logger_id=<logger id>.
        - To clear all data for a logger: call with verb="clear", logger_id=<logger id>.

        Only set the parameters required for the verb you use.

        Args:
            verb: The action to perform. Must be one of: create, delete, clear.
            label: The label for the logger (required for create).
            logger_id: The ID of the logger (required for delete and clear).
            workspace_id: The workspace to associate with the logger (optional for create).
        """
        headers = get_authentication_headers(context)
        if not headers:
            return "The api key was not set"

        if verb == "create":
            if not label:
                return "Missing label for logger creation."
            res = ptt.api.create_http_logger(label, workspace_id, headers=headers)
        elif verb == "delete":
            if not logger_id:
                return "Missing logger_id for deletion."
            res = ptt.api.delete_logger(logger_id, headers=headers)
        elif verb == "clear":
            if not logger_id:
                return "Missing logger_id for clearing data."
            res = ptt.api.clear_data_for_logger(logger_id, headers=headers)
        else:
            return f"Unknown verb: {verb}. Use one of: create, delete, clear."
        return json.dumps(res.json())

    ### Workspaces
    @mcp.tool()
    async def workspace(
        verb: Literal["create", "delete", "edit", "list", "get"],
        context: Context,
        workspace_id: int = None,
        name: str = None,
        description: str = None,
    ) -> str:
        """
        Everything you need to manage workspaces.

        Usage:
        - To create a workspace: call with verb="create", name="workspace name", description="optional description".
        - To delete a workspace: call with verb="delete", workspace_id=<id>. You will need to get the id beforehand.
        - To edit a workspace: call with verb="edit", workspace_id=<id>, name="new name" (optional), description="new description" (optional).
        - To get all workspaces: call with verb="list".
        - To get a workspace by id: call with verb="get", workspace_id=<id>.

        Only set the parameters required for the verb you use.
        """
        headers = get_authentication_headers(context)
        if not headers:
            return "The api key was not set"

        if verb == "create":
            if not name:
                return "Missing workspace name for creation."
            res = ptt.api.create_workspace(name, description or "", headers=headers)
        elif verb == "delete":
            if not workspace_id:
                return "Missing workspace_id for deletion."
            res = ptt.api.delete_workspace(workspace_id, headers=headers)
        elif verb == "edit":
            if not workspace_id:
                return "Missing workspace_id for editing."
            res = ptt.api.edit_workspace(workspace_id, name, description, headers=headers)
        elif verb == "list":
            res = ptt.api.get_workspaces(headers=headers)
        elif verb == "get":
            if not workspace_id:
                return "Missing workspace_id for get."
            res = ptt.api.get_workspace_by_id(workspace_id, headers=headers)
        else:
            return f"Unknown verb: {verb}. Use one of: create, delete, edit, list, get."
        return json.dumps(res.json())

    ### Findings
    @mcp.tool()
    async def findings(
        verb: Literal["get_finding_information_by_id", "get"],
        context: Context,
        finding_id: int = None,
        workspace_id: int = None,
        target_id: int = None,
        task_id: int = None,
        group_duplicates: bool = False,
    ) -> str:
        """
        Manage findings in Pentest-Tools.com.

        Usage:
        - To get all findings: call with verb="get".
        - To get finding information by id: call with verb="get_finding_information_by_id"

        Only set the parameters required for the verb you use.

        Args:
            verb: The action to perform. Must be one of: get_finding_information_by_id, get.
            finding_id: The id of the finding you want more information for.
            workspace_id: (optional) ID of the workspace to show findings from.
            target_id: (optional) ID of the target to show findings from.
            task_id: (optional) ID of the scan to show findings from.
            group_duplicates: (optional) Group duplicates to show only the latest finding from each group. Default is False.
            context: The MCP context object.
        """
        headers = get_authentication_headers(context)
        if not headers:
            return "The api key was not set"

        if verb == "get_finding_information_by_id":
            if not finding_id:
                return "Missing finding_id for get_finding_information_by_id."
            res = ptt.api.get_finding_information_by_id(finding_id=finding_id, headers=headers)
            return json.dumps(res.json())
        elif verb == "get":
            res = ptt.api.get_findings(
                workspace_id=workspace_id,
                target_id=target_id,
                task_id=task_id,
                group_duplicates=group_duplicates,
                headers=headers,
            )
            return json.dumps(res.json())
        else:
            return f"Unknown verb: {verb}. Use one of: get_finding_information_by_id, get."

    @mcp.tool()
    def import_findings(
        findings: list[Finding],
        context: Context,
    ):
        """Import multiple findings into Pentest-Tools.com.

        This is the only way to create new findings manually. Use this when you need to add findings that do not come from scans.

        Args:
            findings: A list of findings.
        """
        headers = get_authentication_headers(context)
        if not headers:
            return "The api key was not set"

        res = ptt.api.import_findings(
            findings=findings,
            headers=headers,
        )
        return json.dumps(res.json())


def entrypoint(mcp_method: str = "stdio"):
    """Entrypoint for ptt-mcp"""
    global MCP_METHOD

    logging.basicConfig(
        level=logging.DEBUG,
        format="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
    )

    if FastMCP is None:
        logging.error(
            "PentestTools has been installed without MCP support. Install with pip install PentestTools[mcp] if you want MCP support."
        )
        return

    logging.info("Starting the PentestTools MCP server")
    if mcp_method == "stdio":
        logging.info("Using the `stdio` mcp type")
        MCP_METHOD = "stdio"

        # Get the api key for stdio
        if not getattr(ptt.api, "api_key", False):
            api_key = os.getenv("PTT_API_KEY", "")
            if not api_key:
                sys.exit(
                    "API key not found. Please provide it using the --key flag or set the PTT_API_KEY environment variable."
                )

            ptt.api.api_key = api_key

        ptt.api.init()  # This call sets the given key

        return mcp.run(transport="stdio")
    elif mcp_method == "http":
        logging.info("Using the `http` mcp type")
        MCP_METHOD = "http"
        return mcp.run(transport="streamable-http")



"""
Sample client for the v2 API of Pentest-Tools.com.
This client starts a Web Server Scan, queries the output and writes the report in a HTML and a PDF file.
A valid API key is necessary for this program to work.

This client contains sample requests for most API methods

API Reference: https://pentest-tools.com/docs/api/v2
"""

import json
import sys
import time
import traceback
import urllib
import os
import pathlib

import requests
from typing import Literal, Union
from typing_extensions import TypedDict

USER_AGENT = os.environ.get("PTT_USER_AGENT", "ptt-cli")
API_URL = os.environ.get("PTT_API_URL", "https://app.pentest-tools.com/api/v2/")
HEADERS = {
    "Content-Type": "application/json",
    "User-Agent": USER_AGENT,
    "Accept": "*/*",
}
API_KEY = None
KEY_READ = False


def generate_api_key(xdg_config_home: str) -> dict[str, str]:
    """Generate a new freemium api key"""
    pathlib.Path(os.path.join(xdg_config_home)).mkdir(parents=True, exist_ok=True)
    config = os.path.join(xdg_config_home, "config.json")

    url = "https://pentest-tools.com/api/auth/token"
    headers = {
        "User-Agent": USER_AGENT,
        "Accept": "*/*",
    }

    response = requests.post(url, headers=headers, cookies={"credentials": "include"})
    key = response.cookies["ptt.auth.token"]

    # Also store the api key in the config file for later use
    with open(config, "w") as file:
        json.dump({"key": key}, file)

    return {"key": key}


# TODO: this doesn't work because HEADER is used as an immutable object for the functions default parameters. Changing this variable won't affect the default parameters of the function.
def init(regenerate: bool = False) -> None:
    """Init function"""
    global API_KEY
    global KEY_READ

    if regenerate or not API_KEY:
        _home = os.path.expanduser("~")

        xdg_config_home = os.environ.get("XDG_CONFIG_HOME") or os.path.join(_home, ".config")

        app_dir = os.path.join(xdg_config_home, "pentesttools")
        if windows_path := os.environ.get("LOCALAPPDATA"):
            app_dir = os.path.join(windows_path, "PentestTools", "Local")

        config = os.path.join(app_dir, "config.json")

        if not regenerate:
            # Try to read the api key from the config file
            try:
                with open(config, "r") as file:
                    config = json.load(file)
                    KEY_READ = True
            # We didn't find the key, generate the config file
            except FileNotFoundError:
                config = generate_api_key(app_dir)
        else:
            config = generate_api_key(app_dir)

        API_KEY = config["key"]

    HEADERS["Authorization"] = f"Bearer {API_KEY}"


class Tool:
    """Map the tool_id that the API knows to the tool name"""

    SUBDOMAIN_FINDER = 20
    WHOIS_LOOKUP = 30
    PORT_SCANNER = 70
    URL_FUZZER = 90
    VIRTUAL_HOSTS_FINDER = 160
    WEBSITE_SCANNER = 170
    ICMP_PING = 240
    SHARE_POINT_SCANNER = 260
    WORDPRESS_SCANNER = 270
    DRUPAL_SCANNER = 280
    JOOMLA_SCANNER = 290
    WEBSITE_RECON = 310
    NETWORK_SCANNER = 350
    SQLI_EXPLOITER = 380
    DOMAIN_FINDER = 390
    PASSWORD_AUDITOR = 400
    SSL_SCANNER = 450
    SNIPER = 490
    WAF_DETECTOR = 500
    API_SCANNER = 510
    CLOUD_SCANNER = 520
    PEOPLE_HUNTER = 530
    KUBERNETES_SCANNER = 540


# Scans

# Two common ways to start a scan is by using either `target` or `target_id`.

# `target` needs to be a simple URL, like "https://example.org". `target_id` is an
# integer you can get from the `get_targets` method.


# For both, you need:
# - tool_id: ID of the tool you want to use
# - tool_params: Options for the tool
# - target or target_id: The target you want to scan, depending on the chosen method
def start_scan(target, tool_id, tool_params, scan_params=None, api_url=API_URL, headers=HEADERS):
    """Start a scan using the given target name"""
    if scan_params is None:
        scan_params = {}

    data = {
        "tool_id": tool_id,
        "target_name": target,
        "tool_params": tool_params,
        **scan_params,
    }
    return requests.post(api_url + "scans/", headers=headers, json=data)


def start_scan_by_targetid(target_id, tool_id, tool_params, api_url=API_URL, headers=HEADERS):
    """Start a scan using the given target_id"""
    data = {"tool_id": tool_id, "target_id": target_id, "tool_params": tool_params}
    return requests.post(api_url + "scans/", headers=headers, json=data)


###Interacting with scans

# After you started a scan, through either method, you may want to interact with it. These are the most commonly
# used methods for interacting with scans after they have been started.

# A scan is identified by a `scan_id`, which can be obtained from running `GET $API/scans`,
# or `get_scans` from this client.

# You can check the status of a scan using the `get_scan_status` function.

# You can get the JSON output of a scan by calling `get_output` with a suitable `scan_id`.
# The previous feature of getting output in a chosen format has moved to the `start_scan` function, through an URL callback. TODO: example


# You may want to stop a running scan, which you can do with `stop_scan`.
# Should you want to delete a scan entirely, tou can use the `delete_scan` function.
def get_scans(
    workspace_id=None,
    target_id=None,
    status=None,
    limit=None,
    page=None,
    start_time=None,
    api_url=API_URL,
    headers=HEADERS,
):
    """Get a list of scans


            res = ptt.api.get_scans(
            workspace_id=workspace_id,
            target_id=target_id,
            status=status,
            limit=limit,
            page=page,
            start_time=start_time,
            headers=headers,
        )


    Specific parameters:
    - workspace_id  -- when set, only the scans from this workspace will be returned
                        (you can get a list of workspaces by using the `get_workspaces` operation)
    - target_id     -- when set, only the scans run on this target will be returned
                        (use `get_targets` for the target list)
    """
    data = {}
    if workspace_id is not None:
        data["workspace_id"] = workspace_id
    if target_id is not None:
        data["target_id"] = target_id
    if status is not None:
        data["status"] = status
    if limit is not None:
        data["limit"] = limit
    if page is not None:
        data["page"] = page
    if start_time is not None:
        data["start_time"] = start_time
    params = "?" + urllib.parse.urlencode(data)
    return requests.get(api_url + f"scans{params}", headers=headers)


def get_scan_status(scan_id, api_url=API_URL, headers=HEADERS):
    """Get the status of a scan"""
    return requests.get(api_url + f"scans/{scan_id}", headers=headers)


def get_output(scan_id, api_url=API_URL, headers=HEADERS):
    """Get the output of a scan"""
    return requests.get(api_url + f"scans/{scan_id}/output", headers=headers)


def get_raw_output(scan_id, api_url=API_URL, headers=HEADERS):
    """Get the raw output of a scan"""
    return requests.get(api_url + f"scans/{scan_id}/raw", headers=headers)


def stop_scan(scan_id, api_url=API_URL, headers=HEADERS):
    """Stop a running scan"""
    return requests.post(api_url + f"scans/{scan_id}/stop", headers=headers)


def delete_scan(scan_id, api_url=API_URL, headers=HEADERS):
    """Delete a scan"""
    return requests.delete(api_url + f"/scans/{scan_id}", headers=headers)


### Targets

# Although you can interact with targets manually, by inputting the URL everytime, Pentest-Tools offers facilities of working with targets.

# The simples workflow involves three functions: Add a target with `add_target`, get all targets with `get_targets` and get a single target with
# `get_target_by_id`.


# Deleting and updating targets remain, for now, an operation you can only do through the site.
def add_target(name, description="", workspace_id=None, api_url=API_URL, headers=HEADERS):
    """Add a new target

    Specific parameters:
    - name          -- the name of the target (must be a hostname, IP address or URL)
    - description   -- a short description of the target (optional)
    - workspace_id  -- the specific workspace in which to add this target (optional)
    """
    data = {"name": name}
    if len(description) > 0:
        data["description"] = description
    if workspace_id is not None:
        data["workspace_id"] = workspace_id

    return requests.post(api_url + "targets", headers=headers, json=data)


def get_targets(
    workspace_id: int = None,
    target_type: str = None,
    limit: int = None,
    page: int = None,
    api_url=API_URL,
    headers=HEADERS,
):
    """
    Get a list of targets.

    Parameters:
    - workspace_id (int, optional): ID of the workspace to show targets from.
    - type (str, optional): The type of target. Possible values: "hostname", "ip_address", "url".
    - limit (int, optional): The maximum number of targets to return.
    - page (int, optional): The page number to return.

    Returns:
    - Response object from requests.get
    """
    params: dict[str, Union[str, int]] = {}
    if workspace_id is not None:
        params["workspace_id"] = workspace_id
    if target_type is not None:
        params["type"] = target_type
    if limit is not None:
        params["limit"] = limit
    if page is not None:
        params["page"] = page

    return requests.get(api_url + "targets", headers=headers, params=params)


def get_target_by_id(target_id, api_url=API_URL, headers=HEADERS):
    """Update the description of a target

    Specific parameters:
    - target_id     -- the ID of the updated target
    """
    return requests.get(api_url + f"targets/{target_id}", headers=headers)


def delete_target(target_id, api_url=API_URL, headers=HEADERS):
    """Delete a target and its scans"""
    return requests.delete(api_url + f"targets/{target_id}", headers=headers)


### Findings
def get_findings(
    workspace_id: int = None,
    target_id: int = None,
    task_id: int = None,
    group_duplicates: bool = False,
    api_url=API_URL,
    headers=HEADERS,
):
    """Get all the findings.

    Parameters:
    - workspace_id (int, optional): ID of the workspace to show findings from.
    - target_id (int, optional): ID of the target to show findings from.
    - task_id (int, optional): ID of the scan to show findings from.
    - group_duplicates (bool, optional): Group duplicates to show only the latest finding from each group. Default is False.

    Returns:
    - Response object from requests.get
    """
    params = {}
    if workspace_id is not None:
        params["workspace_id"] = workspace_id
    if target_id is not None:
        params["target_id"] = target_id
    if task_id is not None:
        params["task_id"] = task_id
    if group_duplicates:
        params["group_duplicates"] = "true"
    else:
        params["group_duplicates"] = "false"

    return requests.get(api_url + "findings/", headers=headers, params=params)

def get_finding_information_by_id(finding_id, api_url=API_URL, headers=HEADERS):
    """Get finding information by id"""
    return requests.get(api_url + f"findings/{finding_id}", headers=headers)


### Finding templates
def get_finding_templates(
    vuln_id_type=None,
    has_ptt_exploit=None,
    limit=None,
    offset=None,
    api_url=API_URL,
    headers=HEADERS,
):
    """Get finding templates

    Specific parameters:
    - vuln_id_type      -- type of vulnerability ID (optional)
    - has_ptt_exploit   -- filter by exploit availability (optional, True/False)
    - limit             -- max number of results (optional)
    - offset            -- offset for pagination (optional)
    """
    params = {}
    if vuln_id_type is not None:
        params["vuln_id_type"] = vuln_id_type
    if has_ptt_exploit is not None:
        params["has_ptt_exploit"] = str(has_ptt_exploit)
    if limit is not None:
        params["limit"] = str(limit)
    if offset is not None:
        params["offset"] = str(offset)

    return requests.get(api_url + "public/finding_templates", headers=headers, params=params)


### Reports
def get_all_reports(api_url=API_URL, headers=HEADERS):
    """Get all reports"""
    return requests.get(api_url + "reports", headers=headers)


def create_report(
    format: Literal["pdf", "html", "json", "csv", "xlsx", "docx"],
    group_by: Literal["target", "vulnerability"],
    source: Literal["scans", "findings"],
    resources: list[int],
    filters: dict = None,
    webhook_url: str = None,
    api_url=API_URL,
    headers=HEADERS,
):
    """
    Create a report.

    Parameters:
    - format: Report format ('pdf', 'html', 'json', 'csv', 'xlsx', 'docx')
    - group_by: Grouping ('target', 'vulnerability')
    - source: Source ('scans', 'findings')
    - resources: List of scan or finding IDs
    - filters: Optional dict of filter options
    - webhook_url: Optional webhook URL for report delivery

    Returns:
    - Response object from requests.post
    """
    data = {
        "format": format,
        "group_by": group_by,
        "source": source,
        "resources": resources,
    }
    if filters is not None:
        data["filters"] = filters
    if webhook_url is not None:
        data["webhook_url"] = webhook_url

    return requests.post(api_url + "reports", headers=headers, json=data)


def get_report_information_by_id(report_id, api_url=API_URL, headers=HEADERS):
    """Get report information by id"""
    return requests.get(f"{api_url}reports/{report_id}", headers=headers)


def delete_report(report_id, api_url=API_URL, headers=HEADERS):
    """Delete a report by its ID"""
    return requests.delete(f"{api_url}reports/{report_id}", headers=headers)


def download_report(report_id, api_url=API_URL, headers=HEADERS):
    """
    Download a report by its ID.

    Parameters:
    - report_id: The ID of the report to download

    Returns:
    - Response object from requests.get
    """
    return requests.get(f"{api_url}reports/{report_id}/download", headers=headers)


### VPN profiles
def get_vpn_profiles(api_url=API_URL, headers=HEADERS):
    """
    Get all VPN profiles.

    Returns:
    - Response object from requests.get
    """
    return requests.get(api_url + "vpn_profiles", headers=headers)


### Wordlists
def get_wordlists(api_url=API_URL, headers=HEADERS):
    """
    Get all available wordlists.

    Returns:
    - Response object from requests.get
    """
    return requests.get(api_url + "wordlists", headers=headers)


def get_wordlist_information_by_id(wordlist_id, api_url=API_URL, headers=HEADERS):
    """
    Get information about a specific wordlist by its ID.

    Parameters:
    - wordlist_id: The ID of the wordlist

    Returns:
    - Response object from requests.get
    """
    return requests.get(f"{api_url}wordlists/{wordlist_id}", headers=headers)


def delete_wordlist(wordlist_id, api_url=API_URL, headers=HEADERS):
    """
    Delete a wordlist by its ID.

    Parameters:
    - wordlist_id: The ID of the wordlist to delete

    Returns:
    - Response object from requests.delete
    """
    return requests.delete(f"{api_url}wordlists/{wordlist_id}", headers=headers)


def get_wordlists_content_by_id(wordlist_id, api_url=API_URL, headers=HEADERS):
    """
    Get the contents of a specific wordlist by its ID.

    Parameters:
    - wordlist_id: The ID of the wordlist

    Returns:
    - Response object from requests.get
    """
    return requests.get(f"{api_url}wordlists/{wordlist_id}/contents", headers=headers)


### HTTP Loggers
def get_http_loggers(workspace_id=None, api_url=API_URL, headers=HEADERS):
    """
    Get all HTTP loggers.

    Parameters:
    - workspace_id: Optional workspace ID to filter loggers

    Returns:
    - Response object from requests.get
    """
    params = {}
    if workspace_id is not None:
        params["workspace_id"] = workspace_id
    return requests.get(api_url + "http_loggers", headers=headers, params=params)


def create_http_logger(label: str, workspace_id: int = None, api_url=API_URL, headers=HEADERS):
    """
    Create a new HTTP request logger.

    Parameters:
    - label: A unique label for the HTTP request logger (required)
    - workspace_id: ID of the workspace to associate the logger with (optional)

    Returns:
    - Response object from requests.post
    """
    data = {"label": label}
    if workspace_id is not None:
        data["workspace_id"] = workspace_id

    # The API expects application/x-www-form-urlencoded
    form_headers = headers.copy()
    form_headers["Content-Type"] = "application/x-www-form-urlencoded"

    return requests.post(
        api_url + "http_loggers",
        headers=form_headers,
        data=urllib.parse.urlencode(data),
    )


def get_logger_information_by_id(logger_id, api_url=API_URL, headers=HEADERS):
    """
    Get information about a specific HTTP logger by its ID.

    Parameters:
    - logger_id: The ID of the HTTP logger

    Returns:
    - Response object from requests.get
    """
    return requests.get(f"{api_url}http_loggers/{logger_id}", headers=headers)


def delete_logger(logger_id, api_url=API_URL, headers=HEADERS):
    """
    Delete an HTTP logger by its ID.

    Parameters:
    - logger_id: The ID of the HTTP logger to delete

    Returns:
    - Response object from requests.delete
    """
    return requests.delete(f"{api_url}http_loggers/{logger_id}", headers=headers)


def get_data_for_logger(logger_id, api_url=API_URL, headers=HEADERS):
    """
    Get data captured by a specific HTTP logger.

    Parameters:
    - logger_id: The ID of the HTTP logger

    Returns:
    - Response object from requests.get
    """
    return requests.get(f"{api_url}http_loggers/{logger_id}/data", headers=headers)


def clear_data_for_logger(logger_id, api_url=API_URL, headers=HEADERS):
    """
    Clear all data captured by a specific HTTP logger.

    Parameters:
    - logger_id: The ID of the HTTP logger

    Returns:
    - Response object from requests.delete
    """
    return requests.delete(f"{api_url}http_loggers/{logger_id}/data", headers=headers)


### Workspaces
def get_workspaces(api_url=API_URL, headers=HEADERS):
    """
    Get all workspaces.

    Returns:
    - Response object from requests.get
    """
    return requests.get(api_url + "workspaces", headers=headers)


def get_workspace_by_id(workspace_id, api_url=API_URL, headers=HEADERS):
    """
    Get information about a specific workspace by its ID.

    Parameters:
    - workspace_id: The ID of the workspace

    Returns:
    - Response object from requests.get
    """
    return requests.get(f"{api_url}workspaces/{workspace_id}", headers=headers)


def delete_workspace(workspace_id, api_url=API_URL, headers=HEADERS):
    """
    Delete a workspace by its ID.

    Parameters:
    - workspace_id: The ID of the workspace to delete

    Returns:
    - Response object from requests.delete
    """
    return requests.delete(f"{api_url}workspaces/{workspace_id}", headers=headers)


def create_workspace(name: str, description: str = "", api_url=API_URL, headers=HEADERS):
    """
    Create a new workspace.

    Parameters:
    - name: Name of the workspace (required)
    - description: Description of the workspace (optional)

    Returns:
    - Response object from requests.post
    """
    data = {"name": name}
    if description:
        data["description"] = description

    form_headers = headers.copy()
    form_headers["Content-Type"] = "application/x-www-form-urlencoded"

    return requests.post(
        api_url + "workspaces",
        headers=form_headers,
        data=urllib.parse.urlencode(data),
    )


def edit_workspace(
    workspace_id: int,
    name: str = None,
    description: str = None,
    api_url=API_URL,
    headers=HEADERS,
):
    """
    Edit an existing workspace.

    Parameters:
    - workspace_id: ID of the workspace to edit (required)
    - name: New name for the workspace (optional)
    - description: New description for the workspace (optional)

    Returns:
    - Response object from requests.put
    """
    data = {}
    if name is not None:
        data["name"] = name
    if description is not None:
        data["description"] = description

    return requests.put(
        f"{api_url}workspaces/{workspace_id}",
        headers=headers,
        json=data,
    )


class Finding(TypedDict):
    name: str
    source: int
    target_name: str
    test_description: str
    vuln_description: str
    vuln_risk_level: int
    vuln_recommendation: str
    cvss_v3: float
    vuln_evidence: str


def import_findings(
    findings: list[Finding],
    workspace_id: int = None,
    api_url=API_URL,
    headers=HEADERS,
):
    """
    The import_findings API call.

    Parameters:
    - findings: list of FindingDict
    - workspace_id: int (optional)
    """
    data: dict[str, Union[list, int]] = {}
    if workspace_id is not None:
        data["workspace_id"] = workspace_id

    data["findings"] = findings

    return requests.post(
        api_url + "findings/import",
        headers=headers,
        json=data,
    )


if __name__ == "__main__":
    # `tool_params` is specific to the tool
    # Here we do a light scan with the Web Server Scanner
    tool_id = Tool.WEBSITE_SCANNER
    tool_params = {"scan_type": "light"}
    target = "http://demo.pentest-tools.com/webapp/"

    # Start the scan
    res = start_scan(target, tool_id, tool_params)
    try:
        res_json = res.json()
    except requests.exceptions.JSONDecodeError:
        print(traceback.format_exc())
        print(res.text)
        sys.exit(1)

    # Get the new `scan_id`
    if "data" in res_json and "created_id" in res_json["data"]:
        scan_id = res_json["data"]["created_id"]
        print("Started scan %i" % scan_id)
    else:
        print("Scan could not start")
        print(f"Status: {res_json['status']}, message: {res_json['message']}")
        sys.exit(1)

    # Poll periodically to check if the scan is finished
    while True:
        time.sleep(2)

        # Get the status of our scan
        status = get_scan_status(scan_id)
        status_name = status.json()["data"]["status_name"]

        if status_name == "finished":
            print("Scan status: %s" % res_json["data"])
            # Get the HTML report and write it to a file
            print("Getting JSON report")
            res = get_output(scan_id)
            output_json = res.json()

            with open("report.json", "w") as file:
                json.dump(output_json, file)

            print("JSON report written to file")
            break



"""Module that holds the global context class, used by the CLI components."""

from __future__ import annotations
from argparse import Namespace
from dataclasses import dataclass
from typing import TYPE_CHECKING, Optional

if TYPE_CHECKING:
    from . import cli


@dataclass
class Context:
    """Holds the variables that are globally used throughout the components of the package."""

    scan_output: Optional[dict] = None
    summary: Optional[cli.ScanStatus] = None
    _args: Optional[Namespace] = None
    _target: Optional[str] = None

    @property
    def args(self) -> Namespace:
        """Parsed command line arguments."""
        if self._args is None:
            raise AttributeError
        return self._args

    @args.setter
    def args(self, val: Namespace) -> None:
        self._args = val

    @property
    def has_args(self) -> bool:
        """Returns if the command line arguments have been set."""
        return self._args is not None

    @property
    def target(self) -> str:
        """The URL that was scanned or to be scanned."""
        return self._target if self._target is not None else self.args.target

    @target.setter
    def target(self, target: Optional[str]) -> None:
        self._target = target



"""
The CLI module for PentestTools.com. It uses PTT API and formats output.
"""

from __future__ import annotations
import argparse
from concurrent.futures import ThreadPoolExecutor
import dataclasses
import functools
import json
import sys
import time
from typing import Optional
import os

import requests

from . import api
from . import banner
from .context import Context
from .logger_config import logger, configure_logger_for_cli
from .text_format import log_runtime_status, print_summary, print_report


STARTED_SCANS = []
EXIT_SUCCESS = 0
EXIT_FAILURE = 1

FAIL_CHOICES = ["low", "medium", "high", "critical", "none"]
SCAN_TYPES = ["light", "deep"]


class CLIException(Exception):
    """Exception specific for this file."""


class Timer:
    """Holds a start_time variable and a delta we want to achive. Used to check if delta time
    has elapsed since start_time and the remaining time."""

    def __init__(self, delta: float):
        self.delta = delta
        self.start_time = time.perf_counter()

    def set(self):
        """Resets the timer start point with the current time."""
        self.start_time = time.perf_counter()

    def is_done(self) -> bool:
        """Checks if wanted time interval has passed."""
        return time.perf_counter() - self.start_time >= self.delta

    def elapsed(self) -> float:
        """Returns how many seconds have passed since start_time."""
        return time.perf_counter() - self.start_time

    def remaining(self):
        """Returns how many seconds are needed to pass for wanted time interval."""
        return self.delta - (time.perf_counter() - self.start_time)


class WaitingFuture:
    """A replacement class for Futures that submit the task to the thread only after a time has
    passed."""

    def __init__(self, executor, timer, *args, **kwargs):
        self.executor = executor
        self.timer = timer
        self.task_args = args
        self.task_kwargs = kwargs

    def done(self):  # pylint: disable=method-hidden
        if not self.timer.is_done():
            return False
        self.timer.set()

        task = self.executor._parent_submit(*self.task_args, **self.task_kwargs)
        self.done = task.done
        self.result = task.result

        return False


class ThrottledThreadPoolExecutor(ThreadPoolExecutor):
    """A subclass of ThreadPoolExecutor that limits the task submission to a time rate."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.timers = {}
        self._parent_submit = super().submit

    def set_func_limit(self, func, delta: float):
        """Sets a time rate limit for a function that will be submitted to this pool."""
        self.timers[func] = Timer(delta)

    def submit(self, func, *args, **kwargs):
        if func not in self.timers:
            return super().submit(func, *args, **kwargs)
        if self.timers[func].is_done():
            self.timers[func].set()
            return super().submit(func, *args, **kwargs)

        return WaitingFuture(self, self.timers[func], func, *args, **kwargs)


class StartScanException(Exception):
    """Exception thrown when a scan start fails."""

    def __init__(self, info, message="Scan failed to start"):
        message += ": " + json.dumps(info)
        super().__init__(message)
        self.info = info


@dataclasses.dataclass
class ResultSummary:
    """Holds the result_summary data from the API."""

    critical: int
    high: int
    info: int
    low: int
    medium: int

    @staticmethod
    def from_json(data: dict) -> ResultSummary:
        """Converts a JSON-like dict to a ResultSummary object."""
        return ResultSummary(
            data["critical"],
            data["high"],
            data["info"],
            data["low"],
            data["medium"],
        )


@dataclasses.dataclass
class ScanStatus:
    """Holds the scan status data from the API."""

    duration: int
    end_time: Optional[str]
    num_finished_tests: int
    num_tests: int
    progress: int
    result_summary: Optional[ResultSummary]
    status_name: str
    start_time: str
    target_id: int

    @staticmethod
    def from_json(data: dict) -> ScanStatus:
        """Converts a JSON-like dict to a ScanStatus object."""
        result_summary = None
        if "result_summary" in data:
            result_summary = ResultSummary.from_json(data["result_summary"])
        return ScanStatus(
            data["duration"],
            data.get("end_time", None),
            data["num_finished_tests"],
            data["num_tests"],
            data["progress"],
            result_summary,
            data["status_name"],
            data["start_time"],
            data["target_id"],
        )


class Scan:
    """Represents a started scan."""

    def __init__(self, scan_id):
        self.scan_id = scan_id

    def get_output(self) -> dict:
        """Returns the result of the scan."""
        return api.get_output(self.scan_id).json()

    def get_scan_status(self) -> ScanStatus:
        """Returns the status of the scan."""
        ret = api.get_scan_status(self.scan_id).json()
        return ScanStatus.from_json(ret["data"])


def start_scan_instance(target, tool_id, scan_type="light", max_scan_time: int = 10):
    """Starts a scan through the API.

    Returns a Scan object associated to the started scan.
    """
    tool_params = {"scan_type": scan_type}
    scan_params = {"max_scan_time": max_scan_time}

    res = api.start_scan(target, tool_id, tool_params, scan_params)
    res_json = res.json()
    if "data" not in res_json or "created_id" not in res_json["data"]:
        raise StartScanException(res_json)
    scan_id = res_json["data"]["created_id"]
    scan = Scan(scan_id)
    STARTED_SCANS.append(scan)
    return scan


def stop_started_scans():
    """Stops all started scans."""
    for scan in STARTED_SCANS:
        api.stop_scan(scan.scan_id)


def cli_resource_handler(func):
    """CLI decorator used to catch all exceptions. Tries to write the scan results and stop the running
    scans."""

    @functools.wraps(func)
    def wrapper(ctx, *args, **kwargs):
        raised_exc = None
        try:
            return func(ctx, *args, **kwargs)
        except requests.exceptions.RequestException as exc:
            raised_exc = exc
            logger.error(
                "An HTTP error has occured. Stopping scans and writing output. Press Ctrl+C to terminate forcefully.\n",
            )
            write_result(ctx)
            try:
                stop_started_scans()
            except requests.exceptions.RequestException:
                pass
        except KeyboardInterrupt as exc:
            raised_exc = exc
            logger.error(
                "A stop was requested. Stopping scans and writing output. "
                "Press Ctrl+C again to terminate forcefully.\n",
            )
            write_result(ctx)
            stop_started_scans()
        except Exception as exc:
            raised_exc = exc
            logger.error(
                "An error has occured. Stopping scans and writing output. Press Ctrl+C to terminate forcefully.\n"
            )
            write_result(ctx)
            stop_started_scans()
            raise

        if ctx.args.verbose:
            raise raised_exc

        return None

    return wrapper


# Currently not used. Replaced by the WaitingFuture.
def throttled_call(delta=1):
    """Limits calls for max 1/delta per second."""

    def decorator(func):
        timer = Timer(delta)

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            if "nowait" in kwargs and kwargs["nowait"]:
                timer.set()
                return func(*args, **kwargs)

            # Wait until time passes
            while not timer.is_done():
                time.sleep(max(0, timer.remaining() - 0.005))

            timer.set()
            return func(*args, **kwargs)

        return wrapper

    return decorator


def write_result(ctx: Context) -> None:
    """Writes the scan result at stdout or in a file.

    Args:
        ctx: Global context.
        target: The URL obtained from the scan summary.
    """
    if not ctx.has_args:
        return

    if ctx.args.output:
        prev_nocolor = ctx.args.nocolor
        ctx.args.nocolor = True
        try:
            with open(ctx.args.output, "w", encoding="utf8") as file:
                if ctx.args.format == "json":
                    file.write(json.dumps(ctx.scan_output))
                elif ctx.args.format == "text":
                    print_report(ctx, ctx.scan_output, file)
                    print_summary(ctx.summary, ctx.target, file)
                else:
                    raise CLIException("Unhandled output format")
        except OSError:
            logger.error("Could not open/write file: %s\n", ctx.args.output)
        else:
            logger.info("Output written to: %s\n", ctx.args.output)
        finally:
            ctx.args.nocolor = prev_nocolor
    else:
        if ctx.args.format == "json":
            print(json.dumps(ctx.scan_output))
        elif ctx.args.format == "text":
            print_report(ctx, ctx.scan_output, sys.stdout)
            print_summary(ctx.summary, ctx.target, sys.stdout)
        else:
            raise CLIException("Unhandled output format")


def scan_loop(ctx: Context, scan: Scan, sec_per_tick=0.2, scan_status_wait=3, scan_output_wait=60) -> None:
    """Renders a live status of the scan, while talking to the API.
    Stops when the scan status is finished. Handles the requests on a separate thread.

    Args:
        ctx: Global context.
        scan: The started scan.
        sec_per_tick: The number of seconds to wait, until rerendering.
    """

    scan_status = scan.get_scan_status()
    del_prev = False

    pool = ThrottledThreadPoolExecutor(max_workers=2)
    pool.set_func_limit(scan.get_scan_status, scan_status_wait)
    pool.set_func_limit(scan.get_output, scan_output_wait)

    status_task = pool.submit(scan.get_scan_status)
    output_task = pool.submit(scan.get_output)

    while True:
        time.sleep(sec_per_tick)

        if status_task.done():
            scan_status = status_task.result()
            status_task = pool.submit(scan.get_scan_status)

        if output_task.done():
            ctx.scan_output = output_task.result()
            output_task = pool.submit(scan.get_output)

        log_runtime_status(
            ctx,
            scan_status,
            del_prev=del_prev,
        )
        del_prev = del_prev or True
        if scan_status.status_name == "finished":
            break

    log_runtime_status(
        ctx,
        scan_status,
        del_prev=del_prev,
    )


def get_option(args, env_name, arg_name, cast=lambda x: x):
    """Gets the option with the correct precedence.
    Environment variables before command-line argument, before default values.

    This is needed for action.yml, where it's much easier to use env vars."""
    env = os.environ.get(env_name)
    arg = getattr(args, arg_name)
    return cast(env or arg)


def parse_args():
    """Parses the command line arguments.

    Returns an ArgumentParser object.
    """
    parser = argparse.ArgumentParser(
        description="""Command-line utility for PentestTools.com.

        Example usage: ptt -q --key <api_key> run website_scanner https://pentest-ground.com
        """,
        prog="ptt",
        add_help=False,
    )
    parser.add_argument(
        "-h",
        "--help",
        action="help",
        default=argparse.SUPPRESS,
        help="Show this help message and exit.",
    )

    parser.add_argument(
        "--fail",
        choices=FAIL_CHOICES,
        default="none",
        help="Define failure criteria. `--fail low` fails if even a low vulnerability finding is found.",
    )

    parser.add_argument("--key", help="The API key. If not provided, obtain one automatically.")
    parser.add_argument(
        "--nocolor",
        action="store_true",
        help="If set, don't color the output at stdout and stderr.",
    )
    parser.add_argument("--verbose", action="store_true", help="If set, print debug information.")
    parser.add_argument(
        "-q",
        "--quiet",
        action="store_true",
        help="If set, suppress info data and errors. Only print the final report.",
    )

    # Create the top-level parser
    subparsers = parser.add_subparsers(required=True)

    # Add the model-context protocol subparser
    # It either runs the MCP server, or, if it was not installed, prints the install command and exits
    parser_mcp = subparsers.add_parser("mcp")
    parser_mcp.set_defaults(mcp=True)  # We want to know that `ptt mcp` has been invoked
    parser_mcp.add_argument(
        "--mcp_method",
        choices=["stdio", "http", "sse"],
        default="stdio",
        help="The method to use for MCP communication. Choices: stdio (default), http, sse.",
    )

    # Create the parser for the first subcommand
    parser_run = subparsers.add_parser("run")
    subparsers_run = parser_run.add_subparsers(dest="website_scanner", required=True)

    # Create the parser for the second subcommand
    parser_website_scanner = subparsers_run.add_parser("website_scanner", add_help=False)

    # Define the final subparser for target and arguments
    parser_website_scanner.add_argument("target", nargs="?", help="The URL to scan.")
    parser_website_scanner.add_argument(
        "--scan_type",
        choices=SCAN_TYPES,
        default="light",
        help="The type of the scan",
    )
    parser_website_scanner.add_argument(
        "--max_scan_time",
        default=10,
        type=int,
        help="Maximum scan duration",
    )

    parser_website_scanner.add_argument(
        "-h",
        "--help",
        action="help",
        default=argparse.SUPPRESS,
        help="Show this help message and exit.",
    )
    parser_website_scanner.add_argument(
        "-o",
        "--output",
        help="File to write the report. If specified, suppress the default report output and redirect to the path.",
    )
    parser_website_scanner.add_argument(
        "--format",
        choices=["text", "json"],
        default="text",
        help="Specify the output format of the vulnerability report.",
    )

    return parser.parse_args()


def transform_args(args: argparse.Namespace) -> None:
    """Transforms some of the arguments given by the user, in a format accepted by the API.

    Args:
        args: The parsed command line arguments.
    """
    args.key = get_option(args, "PTT_API_KEY", "key")

    # Check for MCP.
    if getattr(args, "mcp", False) is True:
        # `ptt mcp` invoked, return early because we don't have to parse the other args, just the type of mcp requested
        args.mcp_method = get_option(args, "PTT_MCP_METHOD", "mcp_method", lambda x: x)
        return
    else:
        # Assigning so we don't have to check with `getattr` again
        args.mcp = False

    args.quiet = get_option(args, "PTT_QUIET", "quiet", bool)
    args.format = get_option(args, "PTT_FORMAT", "format", lambda x: x if x in ["text", "json"] else "text")
    args.target = get_option(args, "PTT_TARGET", "target")
    args.fail = get_option(args, "PTT_FAIL", "fail")
    args.scan_type = get_option(args, "PTT_SCAN_TYPE", "scan_type")
    args.max_scan_time = int(get_option(args, "PTT_MAX_SCAN_TIME", "max_scan_time"))

    if args.fail not in FAIL_CHOICES:
        sys.exit(
            f"error: the failure criteria is not correct. Accepted values are {FAIL_CHOICES}, either through PTT_FAIL or --fail <choice>"
        )

    if args.scan_type not in SCAN_TYPES:
        sys.exit(
            f"error: the scan type is not correct. Accepted values are {SCAN_TYPES}, either through PTT_SCAN_TYPE or --scan_type <choice>"
        )

    if not args.target:
        sys.exit(
            "error: the `target` is required, either through the env var PTT_TARGET or through the command-line: ptt run website_scanner <target>"
        )

    if (not args.target.startswith("http://") or not args.target.startswith("https://")) and "://" not in args.target[
        :10
    ]:
        args.target = f"http://{args.target}"


def validate_args(args: argparse.Namespace):
    """Checks the corectness of the CLI arguments.

    Args:
        args: The parsed command line arguments.

    Returns:
        True, if the arguments are valid, False otherwise.
    """
    if args.output:
        try:
            with open(args.output, "a", encoding="utf-8") as file:
                if not file.writable():
                    logger.error("File is not writable: %s\n", args.output)
                    return False
        except OSError:
            logger.error("Could not open file: %s\n", args.output)
            return False

    return True


@cli_resource_handler
def cli(ctx: Context, scan_loop_kwargs={}):
    """Handles command line arguments, starts the scan, logs the status while
    running, then outputs the vulnerability report."""
    ctx.args = parse_args()
    transform_args(ctx.args)

    # `ptt mcp` was called
    if ctx.args.mcp is True:
        if ctx.args.key:
            api.api_key = ctx.args.key

        from .mcp import entrypoint as mcp_entrypoint

        sys.exit(mcp_entrypoint(mcp_method=ctx.args.mcp_method))

    configure_logger_for_cli(ctx)
    if not validate_args(ctx.args):
        return 1

    manual_key_flag = False
    first_key_try_flag = True
    regenerate = False

    if ctx.args.key:
        api.API_KEY = ctx.args.key
        api.KEY_READ = False
        manual_key_flag = True

    for _ in range(2):
        api.init(regenerate)
        # After this point api.API_KEY is populated

        # Validate api key
        api_key_valid = False
        res = api.get_scan_status(0)
        if res.headers.get("content-type") == "application/json":
            data = res.json()
            if res.status_code == 404 and data["status"] == 404 and "not exist" in data["message"]:
                api_key_valid = True
                break
            elif res.status_code == 200 and "id" in data["data"]:
                api_key_valid = True
                break
        if not api_key_valid:
            # If the key was inserted manually, fail with a log.
            if manual_key_flag:
                logger.error(
                    "The API %s didn't respond properly. Your API key might be invalid.\n",
                    {api.API_URL},
                )
                return EXIT_FAILURE
            # If the key was read from the config, regenerate it.
            elif api.KEY_READ and first_key_try_flag:
                logger.info("API key may be old, regenerating\n")
                regenerate = True
                first_key_try_flag = False
                continue
            # If the key was read from the config and it still fails, fail with a log.
            elif api.KEY_READ and not first_key_try_flag:
                logger.error(
                    "The API %s didn't respond properly on the second retry.\n",
                    {api.API_URL},
                )
                return EXIT_FAILURE

            return EXIT_FAILURE

    logger.error("%s\n", banner.BANNER)
    logger.error("Scanning target: %s\n\n", ctx.args.target)

    try:
        ws = start_scan_instance(
            ctx.args.target,
            api.Tool.WEBSITE_SCANNER,
            ctx.args.scan_type,
            ctx.args.max_scan_time,
        )
    except StartScanException as exc:
        if "message" not in exc.info:
            raise exc
        if (
            "Target URL redirects to" in exc.info["message"]
            or "Target URL is not accessible (DNS error)" in exc.info["message"]
            or "Your current plan only allows" in exc.info["message"]
        ):
            logger.error("%s\n", exc.info["message"])
            sys.exit(1)

        raise exc

    if not isinstance(ws, Scan):
        logger.error("Scan info: %s\n", ws)
        return EXIT_FAILURE

    scan_loop(ctx, ws, **scan_loop_kwargs)
    ctx.scan_output = ws.get_output()
    ctx.summary = ws.get_scan_status()
    target_info = api.get_target_by_id(ctx.summary.target_id)
    ctx.target = target_info.json()["data"]["name"]
    write_result(ctx)

    # Fails if our failure criteria is found within the summary
    result_summary = ctx.summary.result_summary
    if result_summary is not None:
        if ctx.args.fail == "low":
            if any(result_summary.low, result_summary.medium, result_summary.high, result_summary.critical):
                print(
                    "`ptt` scan failed, because it ran with `--fail low` and the report has vulnerabilities with higher or equal risk.",
                    file=sys.stderr,
                )
                return EXIT_FAILURE
        elif ctx.args.fail == "medium":
            if any(result_summary.medium, result_summary.high, result_summary.critical):
                print(
                    "`ptt` scan failed, because it ran with `--fail medium` and the report has vulnerabilities with higher or equal risk.",
                    file=sys.stderr,
                )
                return EXIT_FAILURE
        elif ctx.args.fail == "high":
            if any(result_summary.high, result_summary.critical):
                print(
                    "`ptt` scan failed, because it ran with `--fail high` and the report has vulnerabilities with higher or equal risk.",
                    file=sys.stderr,
                )
                return EXIT_FAILURE
        elif ctx.args.fail == "critical":
            if result_summary.critical:
                print(
                    "`ptt` scan failed, because it ran with `--fail critical` and the report has vulnerabilities with higher or equal risk.",
                    file=sys.stderr,
                )
                return EXIT_FAILURE

    return EXIT_SUCCESS


def entrypoint():
    """Entrypoint function for the CLI."""
    ctx = Context()
    return cli(ctx)



import sys
import logging

from .context import Context

# Guide for logging: https://docs.python.org/3/howto/logging.html#configuring-logging-for-a-library
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
logger.addHandler(logging.NullHandler())


def configure_logger_for_cli(ctx: Context) -> None:
    """Sets the logger for the current module. Logs DEBUG and INFO to stdout and everything from
    WARNING and above, to stderr."""
    stdout_handler = logging.StreamHandler(stream=sys.stdout)
    stdout_handler.setLevel(logging.INFO)
    stdout_handler.addFilter(lambda record: record.levelno == logging.INFO)
    stdout_handler.terminator = ""

    stderr_handler = logging.StreamHandler(stream=sys.stderr)
    stderr_handler.setLevel(logging.WARNING)
    stderr_handler.terminator = ""

    if not ctx.args.quiet:
        logger.addHandler(stdout_handler)
        logger.addHandler(stderr_handler)

    logger.setLevel(logging.INFO)



import sys
import tempfile
import pytest

import responses
import requests
from responses import matchers
from requests.exceptions import RequestException

from pentesttools import cli, api, context
from pentesttools.api import API_URL


API_KEY = "random1vjMsyOuDDYUzeJMQARVJdMcHt"


def mock_start_scan():
    header_match = [
        matchers.header_matcher(
            {
                "Authorization": f"Bearer {API_KEY}",
                "Content-Type": f"application/json",
            }
        )
    ]
    start_match = [
        matchers.json_params_matcher(
            {
                "tool_id": 170,
                "target_name": "http://target.com/1/2?test=me#123",
                "tool_params": {"scan_type": "light"},
            }
        )
    ]

    responses.post(
        url=API_URL + "scans/",
        json={"data": {"created_id": 490, "target_id": 456}},
        status=201,
        match=header_match + start_match,
    )


def mock_scan_status_error():
    header_match = [
        matchers.header_matcher(
            {
                "Authorization": f"Bearer {API_KEY}",
                "Content-Type": f"application/json",
            }
        )
    ]
    status_response = {"status": 404, "message": "The scan with given ID does not exist."}
    responses.get(url=API_URL + "scans/0", json=status_response, status=404, match=header_match)


def mock_scan_status_running():
    header_match = [
        matchers.header_matcher(
            {
                "Authorization": f"Bearer {API_KEY}",
                "Content-Type": f"application/json",
            }
        )
    ]
    status_response = {
        "data": {
            "id": 490,
            "tool_id": 170,
            "target_id": 456,
            "workspace_id": None,
            "status_name": "running",
            "vpn_scan": False,
            "progress": 32,
            "result_summary": {
                "text": "Not found",
                "high": None,
                "medium": None,
                "low": None,
                "info": None,
            },
            "start_time": "2021-02-27 17:03:02",
            "end_time": None,
            "duration": 1900,
            "num_tests": 4,
            "num_finished_tests": 2,
            "scan_original_url": False,
            "tool_params": {
                "scan_type": "light",
                "authentication": None,
                "attack_active": ["xss", "sqli", "ssrf", "http_request_smuggling", "log4j_rce"],
                "attack_passive": ["security_headers", "directory_listing", "sensitive_data"],
                "discovery": ["check_robots", "untrusted_certificates", "interesting_files"],
                "spider": {
                    "approach": "spa",
                    "depth": 10,
                    "exclude_urls": [],
                    "include_urls": [],
                },
                "requests_per_second": 10000,
            },
        }
    }

    for _ in range(10):
        responses.get(url=API_URL + "scans/490", json=status_response, status=200, match=header_match)


def mock_scan_status_finished():
    header_match = [
        matchers.header_matcher(
            {
                "Authorization": f"Bearer {API_KEY}",
                "Content-Type": f"application/json",
            }
        )
    ]
    finished_status_response = {
        "data": {
            "id": 490,
            "tool_id": 170,
            "target_id": 456,
            "workspace_id": None,
            "status_name": "finished",
            "vpn_scan": False,
            "progress": 32,
            "result_summary": {
                "text": "Not found",
                "high": 1,
                "medium": 1,
                "low": 1,
                "info": 2,
            },
            "start_time": "2021-02-27 17:03:02",
            "end_time": "2021-02-27 18:03:02",
            "duration": 1900,
            "num_tests": 20,
            "num_finished_tests": 20,
            "scan_original_url": False,
            "tool_params": {
                "scan_type": "light",
                "authentication": None,
                "attack_active": ["xss", "sqli", "ssrf", "http_request_smuggling", "log4j_rce"],
                "attack_passive": ["security_headers", "directory_listing", "sensitive_data"],
                "discovery": ["check_robots", "untrusted_certificates", "interesting_files"],
                "spider": {
                    "approach": "spa",
                    "depth": 10,
                    "exclude_urls": [],
                    "include_urls": [],
                },
                "requests_per_second": 10000,
            },
        }
    }
    responses.get(url=API_URL + "scans/490", json=finished_status_response, status=200, match=header_match)


def mock_scan_output():
    header_match = [
        matchers.header_matcher(
            {
                "Authorization": f"Bearer {API_KEY}",
                "Content-Type": f"application/json",
            }
        )
    ]
    output_response = {
        "data": {
            "output_type": "finding_list",
            "output_data": {
                "findings": [
                    {
                        "confirmed": True,
                        "cve": ["N/A"],
                        "cvss": None,
                        "id": 857,
                        "name": "Missing security header.",
                        "recommendation": "The Referrer-Policy header shouldbe",
                        "risk_description": "controls how much referrer",
                        "risk_level": 1,
                        "status": "open",
                        "test_description": "Checking Referrer",
                        "test_finished": True,
                        "verified": True,
                        "vuln_description": "Missing security header: Referrer-Policy",
                        "vuln_evidence": {
                            "data": {
                                "headers": ["URL", "Evidence"],
                                "rows": [["http://target.com/", "Response headers do not"]],
                            },
                            "type": "table",
                        },
                        "vuln_id": "WEBSCAN-00-0000029",
                    },
                    {
                        "confirmed": True,
                        "cve": [],
                        "cvss": None,
                        "id": 858,
                        "name": "Communication",
                        "recommendation": "We rcommend",
                        "risk_description": "The communication between",
                        "risk_level": 2,
                        "status": "open",
                        "test_description": "Checking for secure",
                        "test_finished": True,
                        "verified": True,
                        "vuln_description": "Communication is",
                        "vuln_evidence": {
                            "data": {
                                "headers": ["URL", "Evidence"],
                                "rows": [["http://target.ro", "Unencrypted HTTP."]],
                            },
                            "type": "table",
                        },
                        "vuln_id": "WEBSCAN-00-0000005",
                    },
                    {
                        "confirmed": False,
                        "cve": [],
                        "cvss": None,
                        "id": 864,
                        "name": "Vulnerabilities found for server-side software",
                        "recommendation": "We recommend you to upgrade the.",
                        "risk_description": "These expose.",
                        "risk_level": 3,
                        "status": "open",
                        "test_description": "Checking for vulnerabilities ...",
                        "test_finished": True,
                        "verified": False,
                        "vuln_description": "Vulnerabilities found for server-side ",
                        "vuln_evidence": {
                            "data": {
                                "headers": ["Risk Level", "CVSS", "CVE", "Summary", "Affected software"],
                                "rows": [
                                    [
                                        "",
                                        "9.8",
                                        "CVE-2023-25690",
                                        "Some mod_proxy &quot;^/here/(.*)&quot; ",
                                        "http_server 2.4.10",
                                    ],
                                    [
                                        "",
                                        "9",
                                        "CVE-2022-36760",
                                        "Inconsistent Interpretation &apos;",
                                        "http_server 2.4.10",
                                    ],
                                    [
                                        "",
                                        "7.5",
                                        "CVE-2017-3167",
                                        "In Apache httpd 2.2.x before",
                                        "http_server 2.4.10",
                                    ],
                                ],
                            },
                            "type": "table",
                        },
                    },
                    {
                        "confirmed": False,
                        "cve": ["N/A"],
                        "cvss": None,
                        "id": 865,
                        "name": "Nothing was found for",
                        "recommendation": "",
                        "risk_description": "",
                        "risk_level": 0,
                        "status": "open",
                        "test_description": "Checking for client...",
                        "test_finished": True,
                        "verified": False,
                        "vuln_description": "Nothing was found.",
                        "vuln_evidence": {"data": "", "type": "text"},
                        "vuln_id": "NONVULN-00-0000000",
                    },
                    {
                        "confirmed": True,
                        "cve": [],
                        "cvss": None,
                        "id": 867,
                        "name": "Security.txt file",
                        "recommendation": "We recommend allow",
                        "risk_description": "We have detected",
                        "risk_level": 0,
                        "status": "open",
                        "test_description": "Checking for client...",
                        "test_finished": True,
                        "verified": False,
                        "vuln_description": "Missing security.txt",
                        "vuln_evidence": {"data": {"headers": ["URL"], "rows": [["Missing"]]}, "type": "table"},
                        "vuln_id": "WEBSCAN-00-0000059",
                    },
                ]
            },
        }
    }
    responses.get(url=API_URL + "scans/490/output", json=output_response, status=200, match=header_match)


def mock_stop_scan():
    header_match = [
        matchers.header_matcher(
            {
                "Authorization": f"Bearer {API_KEY}",
                "Content-Type": f"application/json",
            }
        )
    ]
    output_response = {}
    responses.post(url=API_URL + "scans/490/stop", json=output_response, status=200, match=header_match)


def mock_get_target_by_id():
    header_match = [
        matchers.header_matcher(
            {
                "Authorization": f"Bearer {API_KEY}",
                "Content-Type": f"application/json",
            }
        )
    ]
    output_response = {"data": {"name": "http://other.target.com"}}
    responses.get(url=API_URL + "targets/456", json=output_response, status=200, match=header_match)


@responses.activate
def test_cli():
    mock_start_scan()
    mock_scan_status_error()
    mock_scan_status_running()
    mock_scan_status_finished()
    mock_scan_output()
    mock_stop_scan()
    mock_get_target_by_id()

    url = "target.com/1/2?test=me#123"
    api.API_KEY = API_KEY
    api.init()
    sys.argv = ["*.py", url, "--nocolor", "--verbose"]
    ret = cli.cli(context.Context(), {"sec_per_tick": 0.01, "scan_status_wait": 0, "scan_output_wait": 0})
    assert ret == 0

    with tempfile.NamedTemporaryFile() as file:
        sys.argv = ["*.py", url, "--nocolor", "--verbose", "--quiet", "--format", "json", "-o", file.name]
        ret = cli.cli(context.Context(), {"sec_per_tick": 0.01, "scan_status_wait": 0, "scan_output_wait": 0})
        assert ret == 0


@responses.activate
def test_cli_fail_status():
    mock_start_scan()
    mock_scan_status_error()

    url = "target.com/1/2?test=me#123"
    api.API_KEY = API_KEY
    api.init()
    sys.argv = ["*.py", url, "--nocolor", "--verbose"]
    with pytest.raises(RequestException):
        cli.cli(context.Context(), {"sec_per_tick": 0.01, "scan_status_wait": 0, "scan_output_wait": 0})



from __future__ import annotations
import html
from typing import TYPE_CHECKING, Optional, TextIO

from .logger_config import logger
from .context import Context
from gettext import ngettext


if TYPE_CHECKING:
    from . import cli


RUNNING_ANIMATION = "|/-\\"
RUN_IDX = 0


def format_time(duration: int):
    if duration < 60:
        return f"{duration} " + ngettext("second   ", "seconds   ", duration)

    duration = duration // 60
    if duration < 60:
        return f"{duration} " + ngettext("minute   ", "minutes   ", duration)

    duration = duration // 60
    return f"{duration} " + ngettext("hour", "hours", duration)


def str_crimson(ctx: Context, text: str) -> str:
    """Turns a string in crimson rgb(151, 50, 50) using ANSI escape sequences."""
    if ctx.args.nocolor:
        return text
    # ANSI 256-color code for crimson-ish (color 131)
    return "\x1b[38;5;131m" + text + "\033[m"


def str_red(ctx: Context, text: str) -> str:
    """Turns a string in red using ANSI escape sequences."""
    if ctx.args.nocolor:
        return text
    return "\x1b[31m" + text + "\033[m"


def str_green(ctx: Context, text: str) -> str:
    """Turns a string in green using ANSI escape sequences."""
    if ctx.args.nocolor:
        return text
    return "\x1b[32m" + text + "\033[m"


def str_orange(ctx: Context, text: str) -> str:
    """Turns a string in orange using ANSI escape sequences."""
    if ctx.args.nocolor:
        return text
    return "\x1b[38;5;214m" + text + "\033[m"


def str_yellow(ctx: Context, text: str) -> str:
    """Turns a string in yellow using ANSI escape sequences."""
    if ctx.args.nocolor:
        return text
    return "\x1b[38;5;226m" + text + "\033[m"


def format_evidence(evidence: dict, prefix: str = "") -> str:
    """Given a JSON of an evidence, generate a string containing the information.

    Args:
        evidence: The JSON obtained from the API.
        prefix: string to append to the beggining of each line.

    Returns the pretty formatted string.
    """
    msg = ""

    if evidence["type"] == "text":
        if not evidence["data"]:
            return msg
        return evidence["data"]

    if evidence["type"] == "table":
        headers = evidence["data"]["headers"]
        rows = evidence["data"]["rows"]
        title_idx = -1
        if "CVE" in headers:
            title_idx = headers.index("CVE")
        elif "Description" in headers:
            title_idx = headers.index("Description")
        elif "Summary" in headers:
            title_idx = headers.index("Summary")

        msg = ""
        for row_idx, row in enumerate(rows, start=1):
            if title_idx == -1:
                msg += prefix + f"- Evidence {row_idx}:\n"
            else:
                msg += (
                    prefix + "- " + html.unescape(row[title_idx])
                    if isinstance(row[title_idx], str)
                    else str(row[title_idx]) + ":\n"
                )

            for idx, (header, data) in enumerate(zip(headers, row)):
                if idx == title_idx:
                    continue

                msg += (
                    prefix + "\t- " + header + ": " + html.unescape(data) if isinstance(data, str) else str(data) + "\n"
                )
            msg += "\n"

    return msg


def format_finding(ctx: Context, index, name, description, risk, recommendation, evidence) -> str:
    """Given finding parameters, generate a string containing the information.

    Returns the pretty formatted string.
    """
    risk_to_severity = ["Info", "Low", "Medium", "High", "Critical"]
    severity = risk_to_severity[risk]

    if severity == "Info":
        msg = str_green(ctx, f"[{index}] {name}")
    elif severity == "Low":
        msg = str_yellow(ctx, f"[{index}] {name}")
    elif severity == "Medium":
        msg = str_orange(ctx, f"[{index}] {name}")
    elif severity == "High":
        msg = str_red(ctx, f"[{index}] {name}")
    else:
        msg = str_crimson(ctx, f"[{index}] {name}")

    if risk == 0 and not recommendation:
        return msg

    msg += "\n"

    if severity == "Info":
        msg += str_green(ctx, f"\t- Risk Level: {risk} ({severity})\n\n")
    elif severity == "Low":
        msg += str_yellow(ctx, f"\t- Risk Level: {risk} ({severity})\n\n")
    elif severity == "Medium":
        msg += str_orange(ctx, f"\t- Risk Level: {risk} ({severity})\n\n")
    elif severity == "High":
        msg += str_red(ctx, f"\t- Risk Level: {risk} ({severity})\n\n")
    else:
        msg += str_crimson(ctx, f"\t- Risk Level: {risk} ({severity})\n\n")

    msg += "\tVulnerability Details:\n"
    evidence_str = format_evidence(evidence, prefix="\t")
    if evidence_str:
        msg += evidence_str + "\n"

    if description != name:
        msg += f"\t- Description: {description}\n"
    if recommendation:
        msg += f"\t- Recommendation: {recommendation}\n"

    msg += "\n"

    return msg


def log_runtime_status(ctx: Context, status: cli.ScanStatus, del_prev=False):
    """Prints the status of a current scan. Shows a bar that fills from 0% to 100% progress. Has
    rotating bar animation."""
    global RUN_IDX

    progress_line = "*" * int((status.progress / 100) * 50) + "_" * int((1 - status.progress / 100) * 50)

    # The extra spaces below are necessary to avoid some formatting duplication

    msg = f"Progress: {progress_line} {status.progress}%  \n"
    if status.status_name == "finished":
        msg += "  Status: Finished    \n"
    else:
        msg += "  Status: " + str_green(ctx, f"Running {RUNNING_ANIMATION[RUN_IDX]}\n")
    RUN_IDX = (RUN_IDX + 1) % len(RUNNING_ANIMATION)
    msg += f"  Start time: {status.start_time}\n"
    msg += f"  Duration: {format_time(status.duration)}\n"
    msg += f"  Total tests: {status.num_tests}\n"
    msg += f"  Finished tests: {status.num_finished_tests}\n"

    if del_prev:
        logger.error("\033\x5bF" * 7)
    logger.error("%s\n", msg)


def print_report(ctx, scan_output, outfile):
    """Prints a the findings of a scan to the given output file. The findings are taken from a JSON
    that was requested previously.

    Args:
        outfile: File to print to. It can also be sys.stdout.
    """
    if scan_output is None:
        return

    data = scan_output["data"]
    if data["output_type"] != "finding_list":
        return
    findings = data["output_data"]["findings"]

    title = "+" + "-" * 58 + "+\n"
    title += "|" + "Vulnerability Scan Report".center(58) + "|\n"
    title += "+" + "-" * 58 + "+\n\n"
    print(title, file=outfile)

    findings.sort(key=lambda x: (-x["risk_level"], not bool(x["recommendation"])))

    for index, finding in enumerate(findings, start=1):
        msg = format_finding(
            ctx,
            index,
            finding["name"],
            finding["vuln_description"],
            finding["risk_level"],
            finding["recommendation"],
            finding["vuln_evidence"],
        )
        print(msg, file=outfile)
    print("", file=outfile)


def print_summary(summary: Optional[cli.ScanStatus], target: str, outfile: TextIO) -> None:
    """Prints a summary table about a finished scan. The summary is a JSON file that was
    requested previously from the API."""
    if summary is None or summary.result_summary is None:
        return

    critical_cnt = summary.result_summary.critical
    high_cnt = summary.result_summary.high
    medium_cnt = summary.result_summary.medium
    low_cnt = summary.result_summary.low
    info_cnt = summary.result_summary.info
    start_time = summary.start_time
    end_time = summary.end_time

    sq_width = 50
    msg = "+" + " TEST summary ".center(sq_width - 2, "-") + "+\n"
    msg += "|" + " " * (sq_width - 2) + "|\n"
    msg += f"|  URL: {target}".ljust(sq_width - 1) + "|\n"
    msg += f"|  Critical Risk Findings: {critical_cnt}".ljust(sq_width - 1) + "|\n"
    msg += f"|  High Risk Findings: {high_cnt}".ljust(sq_width - 1) + "|\n"
    msg += f"|  Medium Risk Findings: {medium_cnt}".ljust(sq_width - 1) + "|\n"
    msg += f"|  Low Risk Findings: {low_cnt}".ljust(sq_width - 1) + "|\n"
    msg += f"|  Info Risk Findings: {info_cnt}".ljust(sq_width - 1) + "|\n"
    msg += f"|  Start time: {start_time}".ljust(sq_width - 1) + "|\n"
    msg += f"|  End time: {end_time}".ljust(sq_width - 1) + "|\n"
    msg += "|" + " " * (sq_width - 2) + "|\n"
    msg += "+" + "-" * (sq_width - 2) + "+\n"

    print(msg, file=outfile)
